import "reflect-metadata";
import { TypeLike } from '../typeslike';
var GenericNode = /** @class */ (function () {
    function GenericNode() {
    }
    GenericNode.prototype.toString = function () {
        var returnStr = this.gType.name + '<';
        var commaStr = '';
        for (var _i = 0, _a = this.gParams; _i < _a.length; _i++) {
            var gParamItem = _a[_i];
            if (gParamItem instanceof GenericNode) {
                returnStr += commaStr + gParamItem.toString();
            }
            else {
                returnStr += commaStr + gParamItem.name;
            }
            commaStr = ', ';
        }
        returnStr += '>';
        return returnStr;
    };
    return GenericNode;
}());
export { GenericNode };
var GenericTokenEnum;
(function (GenericTokenEnum) {
    GenericTokenEnum[GenericTokenEnum["Type"] = 0] = "Type";
    GenericTokenEnum[GenericTokenEnum["Lt"] = 1] = "Lt";
    GenericTokenEnum[GenericTokenEnum["Gt"] = 2] = "Gt";
    GenericTokenEnum[GenericTokenEnum["Comma"] = 3] = "Comma";
})(GenericTokenEnum || (GenericTokenEnum = {}));
var GenericNodeNotNow = /** @class */ (function () {
    function GenericNodeNotNow(_genericNodeCallback) {
        this._genericNodeCallback = _genericNodeCallback;
    }
    Object.defineProperty(GenericNodeNotNow.prototype, "genericNodeCallback", {
        get: function () {
            return this._genericNodeCallback;
        },
        set: function (value) {
            this._genericNodeCallback = value;
        },
        enumerable: true,
        configurable: true
    });
    return GenericNodeNotNow;
}());
export { GenericNodeNotNow };
var GenericTokenizer = /** @class */ (function () {
    function GenericTokenizer() {
        this._tree = null;
        this.currNode = null;
        this.lastToken = null;
        this.ltGtCount = 0;
    }
    Object.defineProperty(GenericTokenizer.prototype, "tree", {
        get: function () {
            if (this.ltGtCount > 0) {
                throw new Error('gt() not found to close lt()');
            }
            return this._tree;
        },
        enumerable: true,
        configurable: true
    });
    GenericTokenizer.create = function () {
        return new GenericTokenizer();
    };
    GenericTokenizer.prototype.tp = function (type) {
        if ((this.lastToken == null)
            || (this.lastToken == GenericTokenEnum.Comma)
            || (this.lastToken == GenericTokenEnum.Lt)) {
            //nothing, OK
        }
        else {
            throw new Error('Invalid token, type(): ' + type);
        }
        if (this._tree == null) {
            this._tree = new GenericNode();
            this._tree.gType = type;
            this._tree.gParams = [];
            this.currNode = this._tree;
        }
        else {
            this.currNode.gParams.push(type);
        }
        this.lastToken = GenericTokenEnum.Type;
        return this;
    };
    GenericTokenizer.prototype.lt = function () {
        if (this.lastToken != GenericTokenEnum.Type) {
            throw new Error('lt() can only come after type()');
        }
        var gParamNew = null;
        if (this.currNode.gParams.length > 0) {
            var gParamActual = this.currNode.gParams.pop();
            if (gParamActual instanceof TypeLike) {
                gParamNew = new GenericNode();
                gParamNew.gType = gParamActual;
                gParamNew.gParent = this.currNode;
                gParamNew.gParams = [];
                this.currNode.gParams.push(gParamNew);
                this.currNode = gParamNew;
            }
        }
        else {
            //does nothing, that's the root type!
        }
        this.ltGtCount++;
        this.lastToken = GenericTokenEnum.Lt;
        return this;
    };
    GenericTokenizer.prototype.gt = function () {
        if (this.ltGtCount == 0) {
            throw new Error('gt() inserted without a preceding lt()');
        }
        if (this.lastToken == GenericTokenEnum.Lt) {
            throw new Error('Generic type "' + this.currNode.gType + '" does not define any parameter type');
        }
        this.currNode = this.currNode.gParent;
        this.ltGtCount--;
        this.lastToken = GenericTokenEnum.Gt;
        return this;
    };
    GenericTokenizer.prototype.comma = function () {
        //nothing
        this.lastToken = GenericTokenEnum.Comma;
        return this;
    };
    GenericTokenizer.resolveNode = function (entityObj, keyItem) {
        var prpGenTypeOrNotNow = Reflect.getMetadata("design:generics", entityObj, keyItem);
        if (prpGenTypeOrNotNow instanceof GenericNode) {
            return prpGenTypeOrNotNow;
        }
        else if (prpGenTypeOrNotNow instanceof GenericNodeNotNow) {
            return prpGenTypeOrNotNow.genericNodeCallback();
        }
        else {
            return null;
        }
    };
    return GenericTokenizer;
}());
export { GenericTokenizer };
//# sourceMappingURL=generic-tokenizer.js.map