import { TypeLike } from '../typeslike';
import { IFieldProcessor } from "../api/field-processor";
import { RecorderLogger, ConsoleLike, RecorderLogLevel, CacheHandler, TypeProcessorEntry, RecorderConfig } from "../api/recorder-config";
import { LazyObservableProvider } from '../api/lazy-observable-provider';
export declare class ConsoleLikeBase implements ConsoleLike {
    private logger;
    private level;
    constructor(logger: RecorderLogger, level: RecorderLogLevel);
    group(...label: any[]): void;
    groupEnd(): void;
    error(message?: any, ...optionalParams: any[]): void;
    warn(message?: any, ...optionalParams: any[]): void;
    log(message?: any, ...optionalParams: any[]): void;
    debug(message?: any, ...optionalParams: any[]): void;
    info(message?: any, ...optionalParams: any[]): void;
    enabledFor(level: RecorderLogLevel): boolean;
    getLevel(): RecorderLogLevel;
}
export declare class RecorderConfigDefault implements RecorderConfig {
    constructor();
    lazyObservableProvider: LazyObservableProvider;
    configLazyObservableProvider(provider: LazyObservableProvider): RecorderConfig;
    private _logLevelMap;
    configLogLevel(logger: RecorderLogger, level: RecorderLogLevel, consoleLike?: ConsoleLike): RecorderConfig;
    getConsole(logger: RecorderLogger): ConsoleLike;
    private _attachPrefix;
    private _cacheStoragePrefix;
    configAttachPrefix(attachPrefix: string): RecorderConfig;
    configCacheStoragePrefix(cacheStoragePrefix: string): RecorderConfig;
    readonly attachPrefix: string;
    readonly cacheStoragePrefix: string;
    private _creationIdName;
    private _playerMetadatasName;
    private _maxLazyRefNotificationPerSecond;
    private _lazyRefNotificationTimeMeasurement;
    private _lazyRefNotificationCountMeasurement;
    private _fieldProcessorEntryMap;
    private _cacheHandler;
    private _maxJsonStringifyForDiagnostic;
    /**
     * Max size of JSON.stringfy for error and log purpose. Default: 300.
     * @param maxJsonStringifyForDiagnostic
     */
    configMaxJsonStringifyForDiagnostic(maxJsonStringifyForDiagnostic: number): RecorderConfig;
    maxJsonStringifyForDiagnostic: number;
    readonly cacheHandler: CacheHandler;
    configCacheHandler(value: CacheHandler): RecorderConfig;
    configCreationIdName(creationIdName: string): RecorderConfig;
    configMetadatasName(playerMetadatasName: string): RecorderConfig;
    configMaxLazyRefNotificationPerSecond(maxLazyRefNotificationPerSecond: number): RecorderConfig;
    configLazyRefNotificationTimeMeasurement(lazyRefNotificationTimeMeasurement: number): RecorderConfig;
    configLazyRefNotificationCountMeasurement(value: number): RecorderConfig;
    creationIdName: string;
    maxLazyRefNotificationPerSecond: number;
    lazyRefNotificationTimeMeasurement: number;
    playerMetadatasName: string;
    lazyRefNotificationCountMeasurement: number;
    configAddFieldProcessors(entries: TypeProcessorEntry<any, any>[]): RecorderConfig;
    getTypeProcessor<L, LM>(type: TypeLike<LM>): IFieldProcessor<L>;
    private _tryReduceLazyRefSubescribersRerun;
    /**
     * Getter tryReduceLazyRefSubescribersRerun
     * @return {boolean}
     */
    readonly tryReduceLazyRefSubescribersRerun: boolean;
    configTryReduceLazyRefSubescribersRerun(tryReduce: boolean): RecorderConfig;
}
