/// <reference types="node" />
import { IFieldProcessor, IFieldProcessorEvents } from '../api/field-processor';
import { TypeLike } from '../typeslike-dev';
import { StringStream, BinaryStream } from '../api/lazy-ref';
import { RecorderDecorators } from '../api/recorder-decorators';
export declare namespace RecorderDecoratorsInternal {
    /**
     * L: In case of LazyRef this is first type parameter of LazyRef.
     */
    interface PropertyOptions<L> {
        persistent: boolean;
        lazyDirectRawRead?: boolean;
        lazyDirectRawWrite?: boolean;
        fieldProcessorResolver?: () => IFieldProcessor<L>;
        /** Framework internal use. */
        fieldProcessorEvents?: IFieldProcessorEvents<L>;
    }
    /**
     * Decorator for get property.
     * \@RecorderDecorators.property() is equivalent to \@RecorderDecorators.property({persistent: true})
     *
     * Examplo:
     * ```ts
       ...
       private _myField: string;
       @RecorderDecorators.property()
       public get myField(): string {
         return this._myField;
       }
       public set myField(value: string) {
         this._myField = value;
       }
       ...
     * ```
     */
    function property<T>(options: PropertyOptions<T>): MethodDecorator;
    function property<T>(): MethodDecorator;
    function playerObjectId(): MethodDecorator;
    /**
     * Decorator for persistent entity.
     *
     * Sample:
     * ```ts
     * ...
     * @JsonPlayback.playerType({playerType: 'org.mypackage.MyPersistentEntity'})
     * export class MyPersistentEntityJs {
     * ...
     * ```
     */
    function playerType(options: RecorderDecorators.playerTypeOptions): ClassDecorator;
    /**
     * Internal use only! It is no a decorator!
     */
    function mountContructorByPlayerTypeMetadataKey(options: RecorderDecorators.playerTypeOptions, entityType: TypeLike<any>): string;
    const DateProcessor: IFieldProcessor<Date>;
    const BufferProcessor: IFieldProcessor<Buffer>;
    const StringProcessor: IFieldProcessor<String>;
    const BinaryStreamProcessor: IFieldProcessor<BinaryStream>;
    const StringStreamProcessor: IFieldProcessor<StringStream>;
}
