import { RecorderConstants } from './recorder-constants';
import { of, from } from 'rxjs';
import { RecorderManagerDefault } from './recorder-manager-default';
import { NonWritableStreamExtraMethods } from '../api/lazy-ref';
import { RecorderLogger, RecorderLogLevel } from '../api/recorder-config';
import { TapeActionType } from '../api/tape';
import { TapeActionDefault } from './tape-default';
import { flatMap, map, share } from 'rxjs/operators';
import streamToObservable from 'stream-to-observable';
import { MemStreamReadableStreamAutoEnd } from './mem-stream-readable-stream-auto-end';
import { LodashLike } from './lodash-like';
export var RecorderDecoratorsInternal;
(function (RecorderDecoratorsInternal) {
    function property() {
        var options = { persistent: true };
        if (arguments.length > 0) {
            options = arguments[0];
        }
        var optionsConst = {
            persistent: true
        };
        if (options) {
            Object.assign(optionsConst, options);
        }
        if (!optionsConst.fieldProcessorEvents) {
            optionsConst.fieldProcessorEvents =
                {};
        }
        var returnFunc = function (target, propertyKey, descriptor) {
            Reflect.defineMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_OBJECT_PROPERTY_OPTIONS, optionsConst, target, propertyKey);
            var oldSet = descriptor.set;
            descriptor.set = function (value) {
                var thisLocal = this;
                var session = LodashLike.get(this, RecorderConstants.ENTITY_SESION_PROPERTY_NAME);
                var consoleLike = session.manager.config.getConsole(RecorderLogger.RecorderDecorators);
                var fieldEtc = RecorderManagerDefault.resolveFieldProcessorPropOptsEtc(session.fielEtcCacheMap, target, propertyKey.toString(), session.manager.config);
                if (fieldEtc.propertyOptions.persistent) {
                    if (consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                        consoleLike.group('JsonPlayback.set' +
                            'propertyOptions.persistent. Intercepting set method for ' + target.constructor.name + '.' + propertyKey + '. target and value:');
                        consoleLike.debug(target);
                        consoleLike.debug(value);
                        consoleLike.groupEnd();
                    }
                    var isOnlazyLoad = LodashLike.get(this, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME);
                    if (value && value.iAmLazyRef) {
                        //nothing
                    }
                    else {
                        if (LodashLike.isObject(target, new Set([Date, Buffer]))) {
                            if (!session) {
                                throw new Error('The property \'' + propertyKey.toString() + '\' of \'' + target.constructor + '\' has a not managed owner. \'' + RecorderConstants.ENTITY_SESION_PROPERTY_NAME + '\' is null or not present');
                            }
                            var actualValue = LodashLike.get(this, propertyKey.toString());
                            if (actualValue !== value) {
                                if (!isOnlazyLoad && !session.isOnRestoreEntireStateFromLiteral()) {
                                    if (!session.isRecording()) {
                                        throw new Error('Invalid operation. It is not recording. is this Error correct?!');
                                    }
                                    if (consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                                        consoleLike.group('JsonPlayback.set' +
                                            '(actualValue !== value) && !isOnlazyLoad && !session.isOnRestoreEntireStateFromLiteral()\n' +
                                            'Recording action: ' + TapeActionType.SetField + '. actual and new value: ');
                                        consoleLike.debug(actualValue);
                                        consoleLike.debug(value);
                                        consoleLike.groupEnd();
                                    }
                                    //do the TapeAction log here
                                    var action_1 = new TapeActionDefault();
                                    action_1.fieldName = propertyKey.toString();
                                    action_1.actionType = TapeActionType.SetField;
                                    var allMD = session.resolveMetadatas({ object: this });
                                    var bMd = allMD.objectMd;
                                    if (bMd.$signature$) {
                                        action_1.ownerSignatureStr = bMd.$signature$;
                                    }
                                    else if (LodashLike.has(this, session.manager.config.creationIdName)) {
                                        action_1.ownerCreationRefId = LodashLike.get(this, session.manager.config.creationIdName);
                                    }
                                    else if (!this._isOnInternalSetLazyObjForCollection) {
                                        throw new Error('The property \'' + propertyKey.toString() + ' of \'' + target.constructor + '\' has a not managed owner');
                                    }
                                    var asyncAddTapeAction = { value: false };
                                    if (value != null && value != undefined) {
                                        var allMD_1 = session.resolveMetadatas({ object: value });
                                        var bMdValue = allMD_1.objectMd;
                                        if (bMdValue.$signature$) {
                                            action_1.settedSignatureStr = bMdValue.$signature$;
                                        }
                                        else if (LodashLike.has(value, session.manager.config.creationIdName)) {
                                            action_1.settedCreationRefId = LodashLike.get(value, session.manager.config.creationIdName);
                                        }
                                        else {
                                            if (LodashLike.isObject(value, new Set([Date, Buffer]))) {
                                                throw new Error('The property \'' + propertyKey.toString() + ' of \'' + this.constructor + '\'. Value can not be anything but primitive in this case. value: ' + value.constructor);
                                            }
                                            action_1.simpleSettedValue = value;
                                        }
                                    }
                                    else {
                                        action_1.simpleSettedValue = null;
                                    }
                                    if (fieldEtc.propertyOptions.lazyDirectRawWrite) {
                                        var processTapeActionAttachRefId$ = session.processTapeActionAttachRefId({ fieldEtc: fieldEtc, value: value, action: action_1, propertyKey: propertyKey.toString() });
                                        //processTapeActionAttachRefId$ = processTapeActionAttachRefId$.pipe(session.addSubscribedObsRxOpr());
                                        asyncAddTapeAction.value = true;
                                        processTapeActionAttachRefId$ = processTapeActionAttachRefId$.pipe(session.registerProvidedObservablesRxOpr(), share());
                                        processTapeActionAttachRefId$.subscribe({
                                            next: function (ptaariValue) {
                                                oldSet.call(thisLocal, ptaariValue.newValue);
                                                if (!ptaariValue.asyncAddTapeAction) {
                                                    session.addTapeAction(action_1);
                                                }
                                            }
                                        });
                                        // action.attachRefId = session.manager.config.cacheStoragePrefix + session.nextMultiPurposeInstanceId();
                                        // if (fieldEtc.fieldProcessorCaller && fieldEtc.fieldProcessorCaller.callToDirectRaw) {
                                        //     let toDirectRaw$ = fieldEtc.fieldProcessorCaller.callToDirectRaw(value, fieldEtc.fieldInfo);
                                        //     toDirectRaw$ = toDirectRaw$.pipe(session.addSubscribedObsRxOpr());
                                        //     asyncAddTapeAction.value = true;
                                        //     toDirectRaw$.subscribe((stream) => {
                                        //         if (stream) {
                                        //             let putOnCache$ = session.manager.config.cacheHandler.putOnCache(action.attachRefId, stream)
                                        //             putOnCache$ = putOnCache$.pipe(session.addSubscribedObsRxOpr());
                                        //             putOnCache$.subscribe(() => {
                                        //                 session.addTapeAction(action);
                                        //             });
                                        //             let getFromCache$ = session.manager.config.cacheHandler.getFromCache(action.attachRefId);
                                        //             getFromCache$ = getFromCache$.pipe(session.addSubscribedObsRxOpr());
                                        //             getFromCache$.subscribe((stream) => {
                                        //                 oldSet.call(this, stream);
                                        //             });
                                        //         } else {
                                        //             if (value) {
                                        //                 throw new Error('The property \'' + propertyKey.toString() + ' of \'' + this.constructor + '\'. NodeJS.ReadableStream is null but value is not null. value: ' + value.constructor);
                                        //             }
                                        //             action.simpleSettedValue = null;
                                        //             action.attachRefId = null;
                                        //             session.addTapeAction(action);
                                        //         }
                                        //     });
                                        // } else {
                                        //     if (!((value as any as NodeJS.ReadableStream).addListener && (value as any as NodeJS.ReadableStream).pipe)) {
                                        //         throw new Error('The property \'' + propertyKey.toString() + ' of \'' + this.constructor + '\'. There is no "IFieldProcessor.toDirectRaw" defined and value is not a NodeJS.ReadableStream. value: ' + value.constructor);
                                        //     } else {
                                        //         let putOnCache$ = session.manager.config.cacheHandler.putOnCache(action.attachRefId, value as any as NodeJS.ReadableStream);
                                        //         putOnCache$ = putOnCache$.pipe(session.addSubscribedObsRxOpr());
                                        //         asyncAddTapeAction.value = true;
                                        //         putOnCache$.subscribe(() => {
                                        //             session.addTapeAction(action);
                                        //         });
                                        //         let getFromCache$ = session.manager.config.cacheHandler.getFromCache(action.attachRefId);
                                        //         getFromCache$ = getFromCache$.pipe(session.addSubscribedObsRxOpr());
                                        //         getFromCache$.subscribe((stream) => {
                                        //             oldSet.call(this, stream);
                                        //         });
                                        //     }
                                        // }
                                    }
                                    else if (fieldEtc.fieldProcessorCaller && fieldEtc.fieldProcessorCaller.callToLiteralValue) {
                                        var processedValue = fieldEtc.fieldProcessorCaller.callToLiteralValue(value, fieldEtc.fieldInfo);
                                        // toLiteralValue$ = toLiteralValue$.pipe(session.addSubscribedObsRxOpr());
                                        asyncAddTapeAction.value = false;
                                        action_1.simpleSettedValue = processedValue;
                                    }
                                    else {
                                        action_1.simpleSettedValue = value;
                                    }
                                    if (!asyncAddTapeAction.value) {
                                        session.addTapeAction(action_1);
                                    }
                                }
                            }
                            else {
                                if (consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                                    consoleLike.group('JsonPlayback.set' +
                                        '(actualValue === value)\n' +
                                        'NOT recording action, BUT may process : ' + TapeActionType.SetField + '. value: ');
                                    consoleLike.debug(value);
                                    consoleLike.groupEnd();
                                }
                            }
                        }
                    }
                    oldSet.call(this, value);
                    if (session && !isOnlazyLoad) {
                        session.notifyAllLazyrefsAboutEntityModification(this, null);
                    }
                }
                else {
                    oldSet.call(this, value);
                    if (consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                        consoleLike.group('JsonPlayback.set' +
                            '!(propertyOptions.persistent && genericNode.gType !== LazyRef && genericNode.gType !== LazyRefPrpMarker). Not intercepting set method for ' + target.constructor.name + '.' + propertyKey + '. target and value:');
                        consoleLike.debug(target);
                        consoleLike.debug(value);
                        consoleLike.groupEnd();
                    }
                }
            };
        };
        return returnFunc;
    }
    RecorderDecoratorsInternal.property = property;
    function playerObjectId() {
        return function (target, propertyKey) {
            var playerObjectIdType = Reflect.getMetadata('design:type', target, propertyKey);
            Reflect.defineMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_OBJECT_ID_TYPE, playerObjectIdType, target);
        };
    }
    RecorderDecoratorsInternal.playerObjectId = playerObjectId;
    /**
     * Decorator for persistent entity.
     *
     * Sample:
     * ```ts
     * ...
     * @JsonPlayback.playerType({playerType: 'org.mypackage.MyPersistentEntity'})
     * export class MyPersistentEntityJs {
     * ...
     * ```
     */
    function playerType(options) {
        return function (target) {
            Reflect.defineMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_TYPE, options, target);
            Reflect.defineMetadata(mountContructorByPlayerTypeMetadataKey(options, target), target, Function);
        };
    }
    RecorderDecoratorsInternal.playerType = playerType;
    /**
     * Internal use only! It is no a decorator!
     */
    function mountContructorByPlayerTypeMetadataKey(options, entityType) {
        return RecorderConstants.REFLECT_METADATA_JSCONTRUCTOR_BY_PLAYER_TYPE_PREFIX +
            entityType.name +
            (options.disambiguationId ? ':' + options.disambiguationId : '') +
            ':' + options.playerType;
    }
    RecorderDecoratorsInternal.mountContructorByPlayerTypeMetadataKey = mountContructorByPlayerTypeMetadataKey;
    // export const BufferProcessor: IFieldProcessor<Buffer> = {
    //     fromLiteralValue: (value, info) => {
    //         if (value) {
    //             return of(Buffer.from(value, 'base64'));
    //         } else {
    //             return of(null);
    //         }
    //     },
    //     toLiteralValue: (value, info) => {
    //         if (value) {
    //             let base64Str = value.toString('base64');                            
    //             return of(base64Str);
    //         }
    //     }
    // };
    // export const StringProcessor: IFieldProcessor<String> = {
    //     fromLiteralValue: (value: string, info: any) => {
    //         return of(value);
    //     },
    //     fromDirectRaw: (stream: NodeJS.ReadableStream, info: any) => {
    //         return from(getStream(stream, {}) as Promise<string>);
    //     }
    // };
    // export const NodeJS.ReadableStreamProcessor: IFieldProcessor<String> = {
    //     fromLiteralValue: (value, info) => {
    //         if (value) {
    //             let base64AB = Buffer.from(value, 'base64');
    //             let ws = new memStreams.WritableStream();
    //             ws.write(base64AB);
    //             let myReadableStreamBuffer = new memStreams.ReadableStream(''); 
    //             myReadableStreamBuffer.push(base64AB);
    //             return of(myReadableStreamBuffer);
    //         } else {
    //             return of(null);
    //         }
    //     },
    //     fromDirectRaw: (stream, info) => {
    //         if (stream) {
    //             if ((stream as NodeJS.ReadableStream).addListener && (stream as NodeJS.ReadableStream).pipe) {
    //                 return of(stream);
    //             } else {
    //                 throw new Error('Not supported');
    //             }
    //         } else {
    //             return of(null);
    //         }
    //     }
    // };
    // export const StringStreamProcessor: IFieldProcessor<StringStream> = {
    //         fromLiteralValue: (value, info) => {
    //             if (value) {
    //                 let base64AB = Buffer.from(value, 'base64');
    //                 let ws = new memStreams.WritableStream();
    //                 ws.write(base64AB);
    //                 let myReadableStreamBuffer = new memStreams.ReadableStream(value); 
    //                 myReadableStreamBuffer.setEncoding('utf-8');
    //                 return of(myReadableStreamBuffer);
    //             } else {
    //                 return of(null);
    //             }
    //         },
    //         fromDirectRaw: (stream, info) => {
    //             if (stream) {
    //                 if ((stream as NodeJS.ReadableStream).addListener && (stream as NodeJS.ReadableStream).pipe) {
    //                     (stream as any as Readable).setEncoding('utf-8');
    //                     return of(stream);
    //                 } else {
    //                     throw new Error('Not supported');
    //                 }
    //             } else {
    //                 return of(null);
    //             }
    //         }
    // };
    RecorderDecoratorsInternal.DateProcessor = {
        fromLiteralValue: function (value) {
            if (value instanceof Number || typeof (value) === 'number') {
                return new Date(value);
            }
            else if (value instanceof String || typeof (value) === 'string') {
                return new Date(value);
            }
            else {
                return null;
            }
        },
        toLiteralValue: function (value) {
            if (value) {
                return value.getTime();
            }
            else {
                return null;
            }
        }
    };
    RecorderDecoratorsInternal.BufferProcessor = {
        fromLiteralValue: function (value) {
            if (value) {
                return Buffer.from(value, 'base64');
            }
            else {
                return null;
            }
        },
        fromDirectRaw: function (respStream$) {
            return respStream$.pipe(flatMap(function (respStream) {
                if (respStream) {
                    var chunkConcatArrRef_1 = { value: [] };
                    return from(streamToObservable(respStream.body)
                        .forEach(function (chunk) {
                        chunkConcatArrRef_1.value.push(chunk);
                    })).pipe(map(function () {
                        return { body: Buffer.concat(chunkConcatArrRef_1.value) };
                    }));
                }
                else {
                    return of({ body: null });
                }
            }));
        },
        toLiteralValue: function (value) {
            if (value) {
                var base64Str = value.toString('base64');
                return of(base64Str);
            }
            else {
                return of(null);
            }
        },
        toDirectRaw: function (value) {
            if (value) {
                // let ws = new memStreams.WritableStream();
                // ws.write(value);
                var myReadableStreamBuffer = new MemStreamReadableStreamAutoEnd('');
                myReadableStreamBuffer.push(value);
                return of({ body: myReadableStreamBuffer });
                // return of(null).pipe(
                //     tap(() => {
                //         myReadableStreamBuffer.emit('end');
                //     }),
                //     map(() => {
                //         return myReadableStreamBuffer;
                //     })
                // );
            }
            else {
                return of({ body: null });
            }
        }
    };
    RecorderDecoratorsInternal.StringProcessor = {
        fromLiteralValue: function (value) {
            return value;
        },
        fromDirectRaw: function (respStream$) {
            return respStream$.pipe(flatMap(function (respStream) {
                if (respStream.body) {
                    respStream.body.setEncoding('utf8');
                    var chunkConcatArrRef_2 = { value: [] };
                    return from(streamToObservable(respStream.body)
                        .forEach(function (chunk) {
                        if (typeof (chunk) === 'string') {
                            chunkConcatArrRef_2.value.push(chunk);
                        }
                        else if (chunk instanceof String) {
                            chunkConcatArrRef_2.value.push(chunk.toString());
                        }
                        else {
                            throw new Error('Not supported!: chunk: ' + chunk);
                        }
                    })).pipe(map(function () {
                        var bufferConc = ''.concat.apply('', chunkConcatArrRef_2.value);
                        return bufferConc;
                    }));
                }
                else {
                    return of(null);
                }
            }));
        },
        toLiteralValue: function (value) {
            return value;
        },
        toDirectRaw: function (value) {
            if (value) {
                var myReadableStreamBuffer = new MemStreamReadableStreamAutoEnd(value.toString());
                myReadableStreamBuffer.setEncoding('utf-8');
                return of({ body: myReadableStreamBuffer });
            }
            else {
                return of({ body: null });
            }
        }
    };
    RecorderDecoratorsInternal.BinaryStreamProcessor = {
        fromLiteralValue: function (value) {
            if (value) {
                var base64AB = Buffer.from(value, 'base64');
                var myReadableStreamBuffer = new MemStreamReadableStreamAutoEnd(base64AB.toString());
                var binaryWRStream = Object.assign(myReadableStreamBuffer, NonWritableStreamExtraMethods);
                return binaryWRStream;
            }
            else {
                return null;
            }
        },
        fromDirectRaw: function (respStream$) {
            return respStream$.pipe(flatMap(function (respStream) {
                if (respStream.body) {
                    if (respStream.body.addListener && respStream.body.pipe) {
                        return of(respStream);
                    }
                    else {
                        throw new Error('Not supported');
                    }
                }
                else {
                    return of({ body: null });
                }
            }));
        },
        toDirectRaw: function (value) {
            if (value) {
                return of({ body: value });
            }
            else {
                return of({ body: null });
            }
        },
        toLiteralValue: function () {
            throw new Error('Not supported!');
        }
    };
    RecorderDecoratorsInternal.StringStreamProcessor = {
        fromLiteralValue: function (value) {
            if (value) {
                // let ws = new memStreams.WritableStream();
                // ws.write(valueBuffer);
                var myReadableStreamBuffer = new MemStreamReadableStreamAutoEnd(value);
                myReadableStreamBuffer.setEncoding('utf-8');
                return myReadableStreamBuffer;
            }
            else {
                return null;
            }
        },
        fromDirectRaw: function (respStream$) {
            return respStream$.pipe(flatMap(function (respStream) {
                if (respStream.body) {
                    if (respStream.body.addListener && respStream.body.pipe) {
                        respStream.body.setEncoding('utf-8');
                        return of(respStream);
                    }
                    else {
                        throw new Error('Not supported');
                    }
                }
                else {
                    return of({ body: null });
                }
            }));
        },
        toDirectRaw: function (value) {
            if (value) {
                value.setEncoding('utf8');
                return of({ body: value });
            }
            else {
                return of({ body: null });
            }
        },
        toLiteralValue: function () {
            throw new Error('Not supported!');
        }
    };
})(RecorderDecoratorsInternal || (RecorderDecoratorsInternal = {}));
//# sourceMappingURL=recorder-decorators-internal.js.map