import { RecorderLogLevel, RecorderLogger } from '../api/recorder-config';
import { RecorderSessionDefault } from './recorder-session-default';
import { RecorderConstants } from './recorder-constants';
import { LazyRefPrpMarker, LazyRefOTMMarker, LazyRefMTOMarker } from '../api/lazy-ref';
import { GenericNode } from '../api/generic-tokenizer';
import { GenericTokenizer } from '../api/generic-tokenizer';
var RecorderManagerDefault = /** @class */ (function () {
    function RecorderManagerDefault(_config) {
        this._config = _config;
        var thisLocal = this;
        if (!this._config.lazyObservableProvider) {
            throw new Error('_httpLazyObservableGen can not be null');
        }
        if (!this._config) {
            throw new Error('_config can not be null');
        }
        thisLocal.consoleLike = this.config.getConsole(RecorderLogger.RecorderManagerDefault);
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
            thisLocal.consoleLike.group('RecorderManagerDefault.constructor()');
            thisLocal.consoleLike.debug(this.config);
            thisLocal.consoleLike.debug(this._config);
            thisLocal.consoleLike.groupEnd();
        }
    }
    RecorderManagerDefault.prototype.createSession = function () {
        var thisLocal = this;
        var result = new RecorderSessionDefault(this);
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
            thisLocal.consoleLike.group('RecorderManagerDefault.createSession():');
            thisLocal.consoleLike.debug(result);
            thisLocal.consoleLike.groupEnd();
        }
        return result;
    };
    Object.defineProperty(RecorderManagerDefault.prototype, "config", {
        get: function () {
            return this._config;
        },
        set: function (value) {
            var thisLocal = this;
            if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
                thisLocal.consoleLike.group('RecorderManagerDefault.config() set: ' + value);
                thisLocal.consoleLike.debug(value);
                thisLocal.consoleLike.groupEnd();
            }
            this._config = value;
        },
        enumerable: true,
        configurable: true
    });
    RecorderManagerDefault.isCollection = function (typeTested) {
        return (typeTested === Array)
            || (typeTested === Set);
    };
    RecorderManagerDefault.resolveFieldProcessorPropOptsEtc = function (fielEtcCacheMap, owner, fieldName, config) {
        if (!fielEtcCacheMap.has(owner)) {
            fielEtcCacheMap.set(owner, new Map());
        }
        if (!fielEtcCacheMap.get(owner).has(fieldName)) {
            var prpType = Reflect.getMetadata('design:type', owner, fieldName);
            var prpGenType = GenericTokenizer.resolveNode(owner, fieldName);
            var lazyLoadedObjType = null;
            var lazyRefMarkerType = void 0;
            var otmCollectionType = null;
            var propertyOptions = Reflect.getMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_OBJECT_PROPERTY_OPTIONS, owner, fieldName);
            var objectIdPrpType = Reflect.getMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_OBJECT_ID_TYPE, owner);
            // let lazyRefGenericParam: TypeLike<any> = null;
            var fieldProcessor = void 0;
            if (propertyOptions && propertyOptions.fieldProcessorResolver) {
                fieldProcessor = propertyOptions.fieldProcessorResolver();
            }
            else {
                if (prpGenType) {
                    fieldProcessor = config.getTypeProcessor(prpGenType.gType);
                }
                if (!fieldProcessor) {
                    fieldProcessor = config.getTypeProcessor(prpType);
                }
                if (!fieldProcessor) {
                    fieldProcessor = {};
                }
            }
            if (prpGenType) {
                if (prpGenType.gParams[0] instanceof GenericNode) {
                    lazyLoadedObjType = prpGenType.gParams[0].gType;
                }
                else {
                    lazyLoadedObjType = prpGenType.gParams[0];
                }
                if (RecorderManagerDefault.isCollection(lazyLoadedObjType)) {
                    otmCollectionType = lazyLoadedObjType;
                    lazyLoadedObjType = prpGenType.gParams[0].gParams[0];
                }
                if (prpGenType.gType === LazyRefOTMMarker
                    || prpGenType.gType === LazyRefMTOMarker
                    || prpGenType.gType === LazyRefPrpMarker) {
                    if (prpGenType.gParams.length <= 0) {
                        throw new Error('LazyRef is not correctly defined: \'' + fieldName + '\' on ' + owner.constructor.name);
                    }
                    lazyRefMarkerType = prpGenType.gType;
                    // if (prpGenType.gParams.length > 0) {
                    // 	if (prpGenType.gParams[0] instanceof GenericNode) {
                    // 		lazyRefGenericParam = (prpGenType.gParams[0] as GenericNode).gType;
                    // 	} else {
                    // 		lazyRefGenericParam = (prpGenType.gParams[0] as TypeLike<any>);
                    // 	}
                    // }
                    if (prpGenType.gType === LazyRefPrpMarker) {
                        if (propertyOptions.fieldProcessorResolver) {
                            fieldProcessor = propertyOptions.fieldProcessorResolver();
                        }
                        else {
                            fieldProcessor = config.getTypeProcessor(lazyLoadedObjType);
                        }
                    }
                }
            }
            var info = {
                fieldName: fieldName,
                fieldType: prpType,
                ownerType: owner.constructor,
                ownerValue: owner
            };
            var fieldProcessorConst_1 = fieldProcessor;
            var propertyOptionsConst_1 = propertyOptions;
            fielEtcCacheMap.get(owner).set(fieldName, {
                prpType: prpType,
                prpGenType: prpGenType,
                objectIdPrpType: objectIdPrpType,
                lazyLoadedObjType: lazyLoadedObjType,
                otmCollectionType: otmCollectionType,
                lazyRefMarkerType: lazyRefMarkerType,
                propertyOptions: propertyOptions,
                fieldProcessorCaller: {
                    callFromLiteralValue: !fieldProcessor || !fieldProcessor.fromLiteralValue ?
                        null :
                        function (value, info) {
                            var fromLiteralValue$ = fieldProcessorConst_1.fromLiteralValue(value, info);
                            //console.log(propertyOptionsConst);
                            if (propertyOptionsConst_1.fieldProcessorEvents
                                && propertyOptionsConst_1.fieldProcessorEvents.onFromLiteralValue) {
                                fromLiteralValue$ = propertyOptionsConst_1.fieldProcessorEvents.onFromLiteralValue(value, info, fromLiteralValue$);
                            }
                            return fromLiteralValue$;
                        },
                    callFromRecordedLiteralValue: !fieldProcessor || !fieldProcessor.fromRecordedLiteralValue ?
                        null :
                        function (value, info) {
                            var fromRecordedLiteralValue$ = fieldProcessorConst_1.fromRecordedLiteralValue(value, info);
                            //console.log(propertyOptionsConst);
                            if (propertyOptionsConst_1.fieldProcessorEvents
                                && propertyOptionsConst_1.fieldProcessorEvents.onFromLiteralValue) {
                                fromRecordedLiteralValue$ = propertyOptionsConst_1.fieldProcessorEvents.onFromRecordedLiteralValue(value, info, fromRecordedLiteralValue$);
                            }
                            return fromRecordedLiteralValue$;
                        },
                    callFromDirectRaw: !fieldProcessor || !fieldProcessor.fromDirectRaw ?
                        null :
                        function (rawResponse, info) {
                            var fromDirectRaw$ = fieldProcessorConst_1.fromDirectRaw(rawResponse, info);
                            if (propertyOptionsConst_1.fieldProcessorEvents
                                && propertyOptionsConst_1.fieldProcessorEvents.onFromDirectRaw) {
                                fromDirectRaw$ = propertyOptionsConst_1.fieldProcessorEvents.onFromDirectRaw(rawResponse, info, fromDirectRaw$);
                            }
                            return fromDirectRaw$;
                        },
                    callToLiteralValue: !fieldProcessor || !fieldProcessor.toLiteralValue ?
                        null :
                        function (value, info) {
                            var toLiteralValue$ = fieldProcessorConst_1.toLiteralValue(value, info);
                            if (propertyOptionsConst_1.fieldProcessorEvents
                                && propertyOptionsConst_1.fieldProcessorEvents.onToLiteralValue) {
                                toLiteralValue$ = propertyOptionsConst_1.fieldProcessorEvents.onToLiteralValue(value, info, toLiteralValue$);
                            }
                            return toLiteralValue$;
                        },
                    callToDirectRaw: !fieldProcessor || !fieldProcessor.toDirectRaw ?
                        null :
                        function (value, info) {
                            var toDirectRaw$ = fieldProcessorConst_1.toDirectRaw(value, info);
                            if (propertyOptionsConst_1.fieldProcessorEvents
                                && propertyOptionsConst_1.fieldProcessorEvents.onToDirectRaw) {
                                toDirectRaw$ = propertyOptionsConst_1.fieldProcessorEvents.onToDirectRaw(value, info, toDirectRaw$);
                            }
                            return toDirectRaw$;
                        }
                },
                // lazyRefGenericParam: lazyRefGenericParam,
                fieldInfo: info
            });
        }
        return fielEtcCacheMap.get(owner).get(fieldName);
    };
    return RecorderManagerDefault;
}());
export { RecorderManagerDefault };
//# sourceMappingURL=recorder-manager-default.js.map