/// <reference types="node" />
import { LazyRef } from '../api/lazy-ref';
import { Observable, OperatorFunction, ObservableInput } from 'rxjs';
import { FieldEtc } from '../api/field-etc';
import { OriginalLiteralValueEntry, PlayerSnapshot, RecorderSessionImplementor } from '../api/recorder-session';
import { TypeLike } from '../typeslike';
import { PlayerMetadatas } from '../api/player-metadatas';
import { RecorderManager } from '../api/recorder-manager';
import { GenericNode } from '../api/generic-tokenizer';
import { LazyRefImplementor } from '../api/lazy-ref';
import { TapeAction, Tape } from '../api/tape';
import { ResponseLike } from '../typeslike';
interface ResolveMetadataReturn {
    refererObjMd: PlayerMetadatas;
    objectMd: PlayerMetadatas;
    playerObjectIdMd: PlayerMetadatas;
    refererObjMdFound: boolean;
    objectMdFound: boolean;
    playerObjectIdMdFound: boolean;
}
export declare class RecorderSessionDefault implements RecorderSessionImplementor {
    private _manager;
    private _objectsBySignature;
    private _objectsByCreationId;
    private _lazyrefsByEntityMap;
    private _fielEtcCacheMap;
    private consoleLike;
    private consoleLikeLogRxOpr;
    private consoleLikeMerge;
    private consoleLikeRestoreState;
    readonly fielEtcCacheMap: Map<Object, Map<String, FieldEtc<any, any>>>;
    private _switchedOffNotificationEntitiesSet;
    private _originalLiteralValueEntries;
    private _nextCreationId;
    private _currentTape;
    private _latestTape;
    private _currentRecordedAtaches;
    private _isOnRestoreEntireStateFromLiteral;
    private _sessionId;
    private _asyncTasksWaitingArrNew;
    private _asyncTasksWaitingSourcesArrNew;
    deepFunctionsSource(obj: any, visitedSet: Set<any>): string;
    testObservableHasCircle(obs: Observable<any>): void;
    logRxOpr<T>(id: string): OperatorFunction<T, T>;
    registerProvidedObservablesRxOpr<T>(): OperatorFunction<T, T>;
    createSerialPendingTasksWaiting(): Observable<void>;
    constructor(_manager: RecorderManager);
    generateEntireStateAsLiteral(): any;
    private restoreEntireStateCallbackTemplate;
    restoreEntireStateFromLiteral(literalState: any): void;
    isOnRestoreEntireStateFromLiteral(): boolean;
    private lazyLoadTemplateCallback;
    private createKeepAllFlagsTemplateCallback;
    private mapKeepAllFlagsRxOprPriv;
    private repeatedValueSet;
    private flatMapKeepAllFlagsRxOprPriv;
    mapJustOnceKeepAllFlagsRxOpr<T, R>(lazyLoadedObj: any, project: (value: T, index?: number) => R): OperatorFunction<T, R>;
    mapKeepAllFlagsRxOpr<T, R>(lazyLoadedObj: any, project: (value: T, index?: number) => R): OperatorFunction<T, R>;
    flatMapJustOnceKeepAllFlagsRxOpr<T, R>(lazyLoadedObj: any, project: (value: T, index?: number) => ObservableInput<R>): OperatorFunction<T, R>;
    flatMapKeepAllFlagsRxOpr<T, R>(lazyLoadedObj: any, project: (value: T, index?: number) => ObservableInput<R>): OperatorFunction<T, R>;
    /**
     * Based on '[TapeAction.java].resolveOwnerValue(IRecorderManager, HashMap<Long, Object>)'
     * @param action
     */
    private actionResolveOwnerValue;
    /**
     * Based on '[TapeAction.java].resolveJavaPropertyName(ObjectMapper, IRecorderManager, HashMap<Long, Object>)'
     * @param action
     */
    private actionResolveFieldName;
    /**
     * Based on '[TapeAction.java].resolveColletion(ObjectMapper, IRecorderManager, HashMap<Long, Object>)'
     * @param action
     */
    private actionResolveColletion;
    /**
     * Based on '[TapeAction.java].resolveSettedValue(ObjectMapper, IRecorderManager, HashMap<Long, Object>)'
     * @param action
     */
    private actionResolveSettedValue;
    /**
     * Based on '[ReplayableDefault.java].replay()'
     */
    private rerunByPlaybacksIgnoreCreateInstance;
    resolveMetadatas(options: {
        object?: any;
        literalObject?: any;
        key?: string;
        refererObject?: Object;
        refererLiteralObject?: any;
        refMap?: Map<Number, any>;
    }): ResolveMetadataReturn;
    createLiteralRefForEntity<T>(realEntity: T): any;
    getEntityInstanceFromLiteralRef<T>(literalRef: any): T;
    /**
     * Getter manager
     * @return {RecorderManager}
     */
    /**
    * Setter manager
    * @param {RecorderManager} value
    */
    manager: RecorderManager;
    processPlayerSnapshot<L>(entityType: TypeLike<L>, playerSnapshot: PlayerSnapshot): L;
    processPlayerSnapshotArray<L>(entityType: TypeLike<L>, playerSnapshot: PlayerSnapshot): Array<L>;
    private newEntityInstanceWithCreationId;
    newEntityInstance<T extends object>(entityType: TypeLike<T>): T;
    startRecording(): void;
    stopRecording(): void;
    recordSave(entity: any): void;
    recordDelete(entity: any): void;
    recordAtache(attach: Observable<NodeJS.ReadableStream>): string;
    storeOriginalLiteralEntry(originalValueEntry: OriginalLiteralValueEntry): void;
    clear(): Observable<void>;
    getLastRecordedTape(): Tape;
    getLastRecordedStreams(): Map<String, Observable<NodeJS.ReadableStream>>;
    getLastRecordedTapeAndStreams(): {
        tape: Tape;
        streams: Map<String, Observable<NodeJS.ReadableStream>>;
    };
    getLastRecordedTapeAsLiteralAndStreams(): {
        tapeLiteral: any;
        streams: Map<String, Observable<NodeJS.ReadableStream>>;
    };
    addTapeAction(action: TapeAction): void;
    isRecording(): boolean;
    getLastRecordedTapeAsLiteral(): any;
    getLastRecordedAtaches(): Map<String, Observable<NodeJS.ReadableStream>>;
    private getPlaybackAsLiteral;
    private getTapeFromLiteral;
    getCachedBySignature<T extends object>(signatureStr: string): T;
    private validatingMetaFieldsExistence;
    validatePlayerSideLiteralObject(literalObject: {}): void;
    validatePlayerSideResponseLike(responseLike: ResponseLike<{} | NodeJS.ReadableStream>): void;
    processWrappedSnapshotFieldArrayInternal<L>(entityType: TypeLike<L>, lazyLoadedColl: any, wrappedSnapshotField: any[]): void;
    processWrappedSnapshotFieldInternal<L>(entityType: TypeLike<L>, snapshotField: any): L;
    private processResultEntityPriv;
    private createLoadedLazyRef;
    tryCacheInstanceBySignature(tryOptions: {
        realInstance: any;
        playerSnapshot: PlayerSnapshot;
        lazySignature?: string;
    }): void;
    protected createNotLoadedLazyRef<L extends object, I>(genericNode: GenericNode, literalLazyObj: any, refMap: Map<Number, any>, refererObj: any, refererKey: string): LazyRef<L, I>;
    private tryGetFromObjectsBySignature;
    createApropriatedLazyRef<L extends object, I>(genericNode: GenericNode, literalLazyObj: any, refererObj: any, refererKey: string, refMap?: Map<Number, any>): LazyRefImplementor<L, I>;
    private metadaKeys;
    private isLiteralObjMetadataKey;
    private removeNonUsedKeysFromLiteral;
    private trySetPlayerObjectIdentifier;
    createCollection(collType: TypeLike<any>, refererObj: any, refererKey: string): any;
    isCollection(typeTested: TypeLike<any>): any;
    addOnCollection(collection: any, element: any): void;
    removeFromCollection(collection: any, element: any): void;
    /**
     * Used exclusively in lazy load or processing literal values from server, it does not create tape actions.
     * @param refMap
     */
    private mergeWithCustomizerPropertyReplection;
    registerEntityAndLazyref(entity: object, lazyRef: LazyRefImplementor<any, any>): void;
    unregisterEntityAndLazyref(entity: object, lazyRef: LazyRefImplementor<any, any>): void;
    notifyAllLazyrefsAboutEntityModification(entity: object, lazyRefSource: LazyRefImplementor<any, any>): void;
    switchOffNotifyAllLazyrefs(entity: object): void;
    switchOnNotifyAllLazyrefs(entity: object): void;
    private _nextMultiPurposeInstanceId;
    /** Framework internal use. */
    nextMultiPurposeInstanceId(): number;
    processTapeActionAttachRefId<T>(options: {
        action: TapeAction;
        fieldEtc: FieldEtc<T, any>;
        value: T;
        propertyKey: string;
    }): Observable<{
        asyncAddTapeAction: boolean;
        newValue: T;
    }>;
    private pendingLazyRefSubscriptionBySign;
    registerLazyRefSubscriptionRxOpr<L>(signature: string): OperatorFunction<L, L>;
    jsonStringfyWithMax(literalObj: any): string;
}
export {};
