import { LazyRefPrpMarker, LazyRefOTMMarker, LazyRefMTOMarker } from '../api/lazy-ref';
import { RecorderManagerDefault } from './recorder-manager-default';
import { map, flatMap, tap, share, timeout } from 'rxjs/operators';
import { of, isObservable } from 'rxjs';
import { RecorderConstants } from './recorder-constants';
import { SetCreator } from './set-creator';
import { JSONHelper } from './json-helper';
import { v1 as uuidv1 } from 'uuid';
import { mapJustOnceRxOpr, combineFirstSerial } from './rxjs-util';
import { GenericNode } from '../api/generic-tokenizer';
import { RecorderDecoratorsInternal } from './recorder-decorators-internal';
import { RecorderLogger, RecorderLogLevel } from '../api/recorder-config';
import { TapeActionType } from '../api/tape';
import { TapeActionDefault, TapeDefault } from './tape-default';
import { LodashLike } from './lodash-like';
import { LazyRefOTMDefault } from './lazy-ref-otm-default';
import { LazyRefMTODefault } from './lazy-ref-mto-default';
import { LazyRefPrpDefault } from './lazy-ref-prp-default';
var UndefinedForMergeAsync = /** @class */ (function () {
    function UndefinedForMergeAsync() {
    }
    UndefinedForMergeAsync.prototype.toString = function () {
        return 'I am an instance of "UndefinedForMerge", just a temporary value before real value from async execution!';
    };
    return UndefinedForMergeAsync;
}());
var DummyUndefinedForMergeAsync = new UndefinedForMergeAsync();
var RecorderSessionDefault = /** @class */ (function () {
    function RecorderSessionDefault(_manager) {
        this._manager = _manager;
        this._objectsBySignature = null;
        this._objectsByCreationId = null;
        this._lazyrefsByEntityMap = null;
        this._fielEtcCacheMap = null;
        this._switchedOffNotificationEntitiesSet = null;
        this._originalLiteralValueEntries = null;
        this._nextCreationId = null;
        this._currentTape = null;
        this._latestTape = null;
        this._currentRecordedAtaches = null;
        this._isOnRestoreEntireStateFromLiteral = false;
        //private _asyncTasksWaitingArr: Set<Observable<any>> = new Set();
        this._asyncTasksWaitingArrNew = new Set();
        this._asyncTasksWaitingSourcesArrNew = new Set();
        this.repeatedValueSet = new Set();
        this._nextMultiPurposeInstanceId = 1;
        this.pendingLazyRefSubscriptionBySign = new Map();
        var thisLocal = this;
        if (!_manager) {
            throw new Error('_manager can not be null');
        }
        thisLocal.consoleLike = _manager.config.getConsole(RecorderLogger.RecorderSessionDefault);
        thisLocal.consoleLikeMerge = _manager.config.getConsole(RecorderLogger.RecorderSessionDefaultMergeWithCustomizerPropertyReplection);
        thisLocal.consoleLikeLogRxOpr = _manager.config.getConsole(RecorderLogger.RecorderSessionDefaultLogRxOpr);
        thisLocal.consoleLikeRestoreState = _manager.config.getConsole(RecorderLogger.RecorderSessionDefaultRestoreState);
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
            thisLocal.consoleLike.group('RecorderSessionDefault.constructor');
            thisLocal.consoleLike.debug(_manager);
            thisLocal.consoleLike.groupEnd();
        }
        this._objectsBySignature = new Map();
        this._objectsByCreationId = new Map();
        this._lazyrefsByEntityMap = new Map();
        this._fielEtcCacheMap = new Map();
        this._switchedOffNotificationEntitiesSet = new Set();
        this._originalLiteralValueEntries = [];
        this._latestTape = [];
        this._sessionId = uuidv1();
    }
    Object.defineProperty(RecorderSessionDefault.prototype, "fielEtcCacheMap", {
        get: function () {
            return this._fielEtcCacheMap;
        },
        enumerable: true,
        configurable: true
    });
    RecorderSessionDefault.prototype.deepFunctionsSource = function (obj, visitedSet) {
        var result = '';
        for (var propt in obj) {
            var prpValue = obj[propt];
            if (typeof (prpValue) === 'object' || typeof (prpValue) === 'function') {
                if (!visitedSet.has(prpValue)) {
                    visitedSet.add(prpValue);
                    if (typeof (prpValue) === 'function') {
                        result += propt + ':\n' + prpValue.toString();
                    }
                    result += '\n' + this.deepFunctionsSource(prpValue, visitedSet);
                }
            }
        }
        return result;
    };
    RecorderSessionDefault.prototype.testObservableHasCircle = function (obs) {
        var visitedSet = new Set();
        var currObs = obs;
        while (currObs && currObs.source) {
            if (visitedSet.has(currObs)) {
                throw new Error('Observable Circle detected:\n' + this.deepFunctionsSource(currObs, new Set()));
            }
            visitedSet.add(obs);
        }
    };
    RecorderSessionDefault.prototype.logRxOpr = function (id) {
        var _this = this;
        var thisLocal = this;
        var resultOpr = function (source) {
            if (thisLocal.consoleLikeLogRxOpr.enabledFor(RecorderLogLevel.Trace)) {
                source.logAllSourceStackRxOprId = 'source observable ' + _this.nextMultiPurposeInstanceId();
                try {
                    throw new Error('logRxOpr(). "Issuer" Stack for id "' + id + '"\n');
                }
                catch (error) {
                    thisLocal.consoleLikeLogRxOpr.debug(error.stack.replace(/^Error: /, ''));
                }
            }
            var result$ = source
                .pipe(map(function (value) {
                if (thisLocal.consoleLikeLogRxOpr.enabledFor(RecorderLogLevel.Trace)) {
                    source.logAllSourceStackRxOprId = 'source observable ' + _this.nextMultiPurposeInstanceId();
                    try {
                        throw new Error('logRxOpr(). "Project" Stack for id "' + id + '"\n');
                    }
                    catch (error) {
                        thisLocal.consoleLikeLogRxOpr.debug(error.stack.replace(/^Error: /, ''));
                    }
                }
                return value;
            }));
            return result$;
        };
        return resultOpr;
    };
    RecorderSessionDefault.prototype.registerProvidedObservablesRxOpr = function () {
        var thisLocal = this;
        //BEGIN: Used to find losted obs.subscribed()
        var stackSubscriberRef = { value: '' };
        try {
            throw new Error('TRACKING');
        }
        catch (error) {
            stackSubscriberRef.value = error.stack;
        }
        //END: Used to find losted obs.subscribed()
        var resultOpr = function (source) {
            var sourceRef = { value: source };
            //thisLocal.testObservableHasCircle(source);
            var result$ = of(null).pipe(tap({
                next: function () {
                    thisLocal._asyncTasksWaitingArrNew.add(result$);
                    thisLocal._asyncTasksWaitingSourcesArrNew.add(sourceRef.value);
                },
                error: function () {
                    thisLocal._asyncTasksWaitingArrNew.add(result$);
                    thisLocal._asyncTasksWaitingSourcesArrNew.add(sourceRef.value);
                }
            }), flatMap(function () {
                sourceRef.value = source;
                return source;
            }), map(function (value) {
                return value;
            }), tap({
                next: function () {
                    thisLocal._asyncTasksWaitingArrNew.delete(result$);
                    thisLocal._asyncTasksWaitingSourcesArrNew.delete(sourceRef.value);
                }
            }));
            return result$;
        };
        return resultOpr;
    };
    RecorderSessionDefault.prototype.createSerialPendingTasksWaiting = function () {
        var thisLocal = this;
        var result$;
        if (thisLocal._asyncTasksWaitingArrNew.size > 0) {
            //result$ = thisLocal.combineFirstSerialPreserveAllFlags(Array.from(thisLocal._asyncTasksWaitingArrNew))
            result$ = combineFirstSerial(Array.from(thisLocal._asyncTasksWaitingArrNew))
                .pipe(thisLocal.logRxOpr('createSerialAsyncTasksWaitingNew'), map(function () { return undefined; }), timeout(3000));
        }
        else {
            result$ = of(null);
        }
        return result$;
    };
    RecorderSessionDefault.prototype.generateEntireStateAsLiteral = function () {
        var _this = this;
        var thisLocal = this;
        var createSerialAsyncTasksWaitingNew$ = this.createSerialPendingTasksWaiting()
            .pipe(map(function () {
            var sessionState = {
                sessionId: _this._sessionId,
                nextCreationId: thisLocal._nextCreationId,
                latestPlaybackArrAsLiteral: [],
                originalLiteralValueEntries: thisLocal._originalLiteralValueEntries
            };
            for (var _i = 0, _a = thisLocal._latestTape; _i < _a.length; _i++) {
                var tapeItem = _a[_i];
                sessionState.latestPlaybackArrAsLiteral.push(thisLocal.getPlaybackAsLiteral(tapeItem));
            }
            if (thisLocal._currentTape) {
                sessionState.currentTapeAsLiteral = thisLocal.getPlaybackAsLiteral(thisLocal._currentTape);
            }
            return sessionState;
        }));
        return createSerialAsyncTasksWaitingNew$;
    };
    RecorderSessionDefault.prototype.restoreEntireStateCallbackTemplate = function (callback) {
        this._isOnRestoreEntireStateFromLiteral = true;
        try {
            return callback();
        }
        finally {
            this._isOnRestoreEntireStateFromLiteral = false;
        }
    };
    RecorderSessionDefault.prototype.restoreEntireStateFromLiteral = function (literalState) {
        var _this = this;
        var thisLocal = this;
        thisLocal.restoreEntireStateCallbackTemplate(function () {
            var literalStateLocal = literalState;
            thisLocal._nextCreationId = literalStateLocal.nextCreationId;
            thisLocal._originalLiteralValueEntries = literalStateLocal.originalLiteralValueEntries;
            if (literalStateLocal.currentTapeAsLiteral) {
                thisLocal._currentTape = thisLocal.getTapeFromLiteral(literalStateLocal.currentTapeAsLiteral);
            }
            else {
                thisLocal._currentTape = null;
            }
            _this._latestTape = [];
            for (var _i = 0, _a = literalStateLocal.latestPlaybackArrAsLiteral; _i < _a.length; _i++) {
                var tapeLiteral = _a[_i];
                thisLocal._latestTape.push(thisLocal.getTapeFromLiteral(tapeLiteral));
            }
            var originalLiteralValueEntriesLengthInitial = thisLocal._originalLiteralValueEntries.length;
            var _loop_1 = function (originalLiteralValueEntry) {
                if (originalLiteralValueEntriesLengthInitial !== thisLocal._originalLiteralValueEntries.length) {
                    throw new Error('There is some error on "this.storeOriginalLiteralEntry()"' +
                        ' manipulation. Initial length ' + originalLiteralValueEntriesLengthInitial +
                        ' is differrent of actual ' + thisLocal._originalLiteralValueEntries.length);
                }
                if (originalLiteralValueEntry.method === 'processResultEntity'
                    || originalLiteralValueEntry.method === 'processResultEntityArray'
                    || originalLiteralValueEntry.method === 'newEntityInstance') {
                    var jsType = Reflect.getMetadata(originalLiteralValueEntry.reflectFunctionMetadataTypeKey, Function);
                    if (!jsType) {
                        throw new Error('the classe \'' + originalLiteralValueEntry.reflectFunctionMetadataTypeKey + ' is not using the decorator \'RecorderDecorators.playerType\'. Entry:\n' + _this.jsonStringfyWithMax(originalLiteralValueEntry));
                    }
                    if (originalLiteralValueEntry.method === 'processResultEntity') {
                        thisLocal.processPlayerSnapshot(jsType, originalLiteralValueEntry.playerSnapshot);
                    }
                    else if (originalLiteralValueEntry.method === 'processResultEntityArray') {
                        thisLocal.processPlayerSnapshotArray(jsType, originalLiteralValueEntry.playerSnapshot);
                    }
                    else if (originalLiteralValueEntry.method === 'newEntityInstance') {
                        thisLocal.newEntityInstanceWithCreationId(jsType, originalLiteralValueEntry.ref.creationId);
                    }
                    else {
                        throw new Error('This should not happen');
                    }
                }
                else if (originalLiteralValueEntry.method === 'lazyRef') {
                    originalLiteralValueEntry.ownerSignatureStr;
                    if (originalLiteralValueEntry.ownerSignatureStr) {
                        if (thisLocal.consoleLikeRestoreState.enabledFor(RecorderLogLevel.Trace)) {
                            thisLocal.consoleLikeRestoreState.debug('RecorderSessionDefault.restoreEntireStateFromLiteral: (ownerSignatureStr): ownerSignatureStr found for original literal value entry, the owner must be a player side component. Entry:\n' + _this.jsonStringfyWithMax(originalLiteralValueEntry));
                        }
                        var ownerEnt = _this._objectsBySignature.get(originalLiteralValueEntry.ownerSignatureStr);
                        if (!ownerEnt) {
                            throw new Error('ownerEnt not found for signature: ' + originalLiteralValueEntry.ownerSignatureStr);
                        }
                        var lazyRef_1 = LodashLike.get(ownerEnt, originalLiteralValueEntry.ownerFieldName);
                        if (!lazyRef_1) {
                            throw new Error('ownerEnt has no field: ' + originalLiteralValueEntry.ownerFieldName);
                        }
                        if (!lazyRef_1.iAmLazyRef) {
                            throw new Error(originalLiteralValueEntry.ownerFieldName + ' is not a LazyRef for ' + ownerEnt);
                        }
                        var processResponseResult = lazyRef_1.processResponse({ body: originalLiteralValueEntry.playerSnapshot });
                        if (isObservable(processResponseResult)) {
                            var processResponseResult$ = processResponseResult.pipe(thisLocal.registerProvidedObservablesRxOpr(), share());
                            processResponseResult$.subscribe(function () {
                                thisLocal.consoleLikeRestoreState.debug('RecorderSessionDefault.restoreEntireStateFromLiteral\n' +
                                    '  -> processResponseResult$.subscribe(): (!ownerEnt): ' +
                                    'No owner entity for original literal value entry, and lazyRef.processResponse() returned Observable. ' +
                                    'This must be an LazyRefPrp marked with lazyDirectRawRead, it meas tgis LazytRef is usink cache.\n' +
                                    'lazyRef:\n' +
                                    lazyRef_1);
                            });
                        }
                    }
                    else {
                        if (thisLocal.consoleLikeRestoreState.enabledFor(RecorderLogLevel.Trace)) {
                            thisLocal.consoleLikeRestoreState.debug('RecorderSessionDefault.restoreEntireStateFromLiteral: (!ownerEnt): ' +
                                'No owner entity for original literal value entry, the owner must be a\n' +
                                'player side component. Doing nothing, in any next literal value entry\n' +
                                'there will exist an action with type \'processResultEntity\' that will\n' +
                                'put the entity on cache. Entry:\n' +
                                _this.jsonStringfyWithMax(originalLiteralValueEntry));
                        }
                    }
                }
                else {
                    throw new Error('This should not happen');
                }
            };
            for (var _b = 0, _c = thisLocal._originalLiteralValueEntries; _b < _c.length; _b++) {
                var originalLiteralValueEntry = _c[_b];
                _loop_1(originalLiteralValueEntry);
            }
            thisLocal.rerunByPlaybacksIgnoreCreateInstance();
        });
    };
    RecorderSessionDefault.prototype.isOnRestoreEntireStateFromLiteral = function () {
        return this._isOnRestoreEntireStateFromLiteral;
    };
    RecorderSessionDefault.prototype.lazyLoadTemplateCallback = function (lazyLoadedObj, originalCb) {
        try {
            LodashLike.set(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME, true);
            return originalCb();
        }
        finally {
            LodashLike.set(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME, false);
        }
    };
    RecorderSessionDefault.prototype.createKeepAllFlagsTemplateCallback = function (lazyLoadedObj) {
        var thisLocal = this;
        var syncIsOn = LodashLike.get(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME);
        var syncIsOn2 = this._isOnRestoreEntireStateFromLiteral;
        return function (originalCb) {
            var asyncIsOn = LodashLike.get(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME);
            var asyncIsOn2 = thisLocal._isOnRestoreEntireStateFromLiteral;
            LodashLike.set(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME, syncIsOn);
            thisLocal._isOnRestoreEntireStateFromLiteral = syncIsOn2;
            try {
                return originalCb();
            }
            finally {
                LodashLike.set(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME, asyncIsOn);
                thisLocal._isOnRestoreEntireStateFromLiteral = asyncIsOn2;
            }
        };
    };
    RecorderSessionDefault.prototype.mapKeepAllFlagsRxOprPriv = function (when, turnOnMode, lazyLoadedObj, project, thisArg) {
        var thisLocal = this;
        var syncIsOn = LodashLike.get(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME);
        var syncIsOn2 = this._isOnRestoreEntireStateFromLiteral;
        var isPipedCallbackDone = { value: false, result: null };
        var newOp = function (source) {
            var projectExtentend = function (value, index) {
                if (!isPipedCallbackDone.value || when === 'eachPipe') {
                    isPipedCallbackDone.value = true;
                    var asyncIsOn = LodashLike.get(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME);
                    var asyncIsOn2 = thisLocal._isOnRestoreEntireStateFromLiteral;
                    if (!turnOnMode || turnOnMode.lazyLoad === 'none') {
                        LodashLike.set(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME, syncIsOn);
                    }
                    else {
                        LodashLike.set(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME, turnOnMode.lazyLoad);
                    }
                    if (!turnOnMode || turnOnMode.restoreStare === 'none') {
                        thisLocal._isOnRestoreEntireStateFromLiteral = syncIsOn2;
                    }
                    else {
                        thisLocal._isOnRestoreEntireStateFromLiteral = turnOnMode.restoreStare;
                    }
                    try {
                        isPipedCallbackDone.result = project(value, index);
                    }
                    finally {
                        LodashLike.set(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME, asyncIsOn);
                        thisLocal._isOnRestoreEntireStateFromLiteral = asyncIsOn2;
                    }
                }
                return isPipedCallbackDone.result;
            };
            return source
                .pipe(map(projectExtentend, thisArg));
        };
        return newOp;
    };
    RecorderSessionDefault.prototype.flatMapKeepAllFlagsRxOprPriv = function (when, turnOnMode, lazyLoadedObj, project, concurrent) {
        var thisLocal = this;
        var syncIsOn = LodashLike.get(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME);
        var syncIsOn2 = this._isOnRestoreEntireStateFromLiteral;
        var isPipedCallbackDone = { value: false, result: null };
        var newOp = function (source) {
            var projectExtentend = function (value, index) {
                if (!isPipedCallbackDone.value || when === 'eachPipe') {
                    isPipedCallbackDone.value = true;
                    var asyncIsOn = LodashLike.get(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME);
                    var asyncIsOn2 = thisLocal._isOnRestoreEntireStateFromLiteral;
                    if (!turnOnMode || turnOnMode.lazyLoad === 'none') {
                        LodashLike.set(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME, syncIsOn);
                    }
                    else {
                        LodashLike.set(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME, turnOnMode.lazyLoad);
                    }
                    if (!turnOnMode || turnOnMode.restoreStare === 'none') {
                        thisLocal._isOnRestoreEntireStateFromLiteral = syncIsOn2;
                    }
                    else {
                        thisLocal._isOnRestoreEntireStateFromLiteral = turnOnMode.restoreStare;
                    }
                    thisLocal._isOnRestoreEntireStateFromLiteral = syncIsOn2;
                    try {
                        isPipedCallbackDone.result = project(value, index);
                    }
                    finally {
                        LodashLike.set(lazyLoadedObj, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME, asyncIsOn);
                        thisLocal._isOnRestoreEntireStateFromLiteral = asyncIsOn2;
                    }
                }
                return isPipedCallbackDone.result;
            };
            return source
                .pipe(flatMap(projectExtentend, concurrent));
        };
        return newOp;
    };
    RecorderSessionDefault.prototype.mapJustOnceKeepAllFlagsRxOpr = function (lazyLoadedObj, project) {
        return this.mapKeepAllFlagsRxOprPriv('justOnce', { lazyLoad: 'none', restoreStare: 'none' }, lazyLoadedObj, project);
    };
    RecorderSessionDefault.prototype.mapKeepAllFlagsRxOpr = function (lazyLoadedObj, project) {
        return this.mapKeepAllFlagsRxOprPriv('eachPipe', { lazyLoad: 'none', restoreStare: 'none' }, lazyLoadedObj, project);
    };
    RecorderSessionDefault.prototype.flatMapJustOnceKeepAllFlagsRxOpr = function (lazyLoadedObj, project) {
        return this.flatMapKeepAllFlagsRxOprPriv('justOnce', { lazyLoad: 'none', restoreStare: 'none' }, lazyLoadedObj, project);
    };
    RecorderSessionDefault.prototype.flatMapKeepAllFlagsRxOpr = function (lazyLoadedObj, project) {
        return this.flatMapKeepAllFlagsRxOprPriv('eachPipe', { lazyLoad: 'none', restoreStare: 'none' }, lazyLoadedObj, project);
    };
    /**
     * Based on '[TapeAction.java].resolveOwnerValue(IRecorderManager, HashMap<Long, Object>)'
     * @param action
     */
    RecorderSessionDefault.prototype.actionResolveOwnerValue = function (action) {
        if (action.ownerSignatureStr) {
            return this._objectsBySignature.get(action.ownerSignatureStr);
        }
        else if (action.ownerCreationId) {
            throw new Error('This should not happen. Action: ' + action.actionType);
        }
        else if (action.ownerCreationRefId) {
            return this._objectsByCreationId.get(action.ownerCreationRefId);
        }
        else {
            throw new Error('This should not happen. Action: ' + this.jsonStringfyWithMax(action));
        }
    };
    /**
     * Based on '[TapeAction.java].resolveJavaPropertyName(ObjectMapper, IRecorderManager, HashMap<Long, Object>)'
     * @param action
     */
    RecorderSessionDefault.prototype.actionResolveFieldName = function (action) {
        return action.fieldName;
    };
    /**
     * Based on '[TapeAction.java].resolveColletion(ObjectMapper, IRecorderManager, HashMap<Long, Object>)'
     * @param action
     */
    RecorderSessionDefault.prototype.actionResolveColletion = function (action) {
        if (action.actionType == TapeActionType.CollectionAdd || action.actionType == TapeActionType.CollectionRemove) {
            try {
                return this.actionResolveOwnerValue(action)[this.actionResolveFieldName(action)];
            }
            catch (e) {
                var newErr = new Error('This should not happen. action. Action ' + this.jsonStringfyWithMax(action));
                newErr.reason = e;
                throw newErr;
            }
        }
        else {
            return null;
        }
    };
    /**
     * Based on '[TapeAction.java].resolveSettedValue(ObjectMapper, IRecorderManager, HashMap<Long, Object>)'
     * @param action
     */
    RecorderSessionDefault.prototype.actionResolveSettedValue = function (action, fieldEtc) {
        var thisLocal = this;
        var resolvedSettedValue;
        if (action.settedCreationRefId) {
            resolvedSettedValue = this._objectsByCreationId.get(action.settedCreationRefId);
        }
        else if (fieldEtc.lazyRefMarkerType === LazyRefPrpMarker) {
            if (!action.attachRefId) {
                if (fieldEtc.fieldProcessorCaller.callFromLiteralValue) {
                    resolvedSettedValue = fieldEtc.fieldProcessorCaller.callFromLiteralValue(action.simpleSettedValue, fieldEtc.fieldInfo);
                }
            }
            else if (action.attachRefId) {
                if (fieldEtc.fieldProcessorCaller.callFromDirectRaw) {
                    var getFromCache$ = thisLocal.manager.config.cacheHandler.getFromCache(action.attachRefId);
                    var callFromDirectRawRL$ = fieldEtc.fieldProcessorCaller.callFromDirectRaw(getFromCache$.pipe(flatMap(function (stream) {
                        return of({
                            body: stream
                        });
                    })), fieldEtc.fieldInfo);
                    resolvedSettedValue = callFromDirectRawRL$;
                }
                else {
                    var getFromCache$ = thisLocal.manager.config.cacheHandler.getFromCache(action.attachRefId);
                    resolvedSettedValue = getFromCache$.pipe(flatMap(function (stream) {
                        return of({
                            body: stream
                        });
                    }));
                }
            }
            else {
                throw new Error('Invalid action. LazyRefPrp invalid values: ' + this.jsonStringfyWithMax(action));
            }
        }
        else if (action.settedSignatureStr) {
            resolvedSettedValue = thisLocal._objectsBySignature.get(action.settedSignatureStr);
        }
        else if (action.fieldName) {
            if (fieldEtc.fieldProcessorCaller.callFromRecordedLiteralValue) {
                resolvedSettedValue =
                    fieldEtc.fieldProcessorCaller.callFromRecordedLiteralValue(action.simpleSettedValue, fieldEtc.fieldInfo);
            }
            else if (fieldEtc.fieldProcessorCaller.callFromLiteralValue) {
                resolvedSettedValue = fieldEtc.fieldProcessorCaller.callFromLiteralValue(action.simpleSettedValue, fieldEtc.fieldInfo);
            }
            else {
                resolvedSettedValue = action.simpleSettedValue;
            }
        }
        return resolvedSettedValue;
    };
    /**
     * Based on '[ReplayableDefault.java].replay()'
     */
    RecorderSessionDefault.prototype.rerunByPlaybacksIgnoreCreateInstance = function () {
        var _this = this;
        var thisLocal = this;
        //let obsArr: Observable<void>[] = [];
        var allPlaybacks = this._latestTape.slice().concat((this._currentTape ? [this._currentTape] : []));
        for (var _i = 0, allPlaybacks_1 = allPlaybacks; _i < allPlaybacks_1.length; _i++) {
            var tape = allPlaybacks_1[_i];
            var _loop_2 = function (action) {
                if (action.actionType != TapeActionType.Create) {
                    var resolvedOwnerValue_1 = thisLocal.actionResolveOwnerValue(action);
                    var resolvedFieldName_1 = thisLocal.actionResolveFieldName(action);
                    var resolvedCollection_1 = thisLocal.actionResolveColletion(action);
                    var fieldEtc_1 = RecorderManagerDefault.resolveFieldProcessorPropOptsEtc(thisLocal.fielEtcCacheMap, resolvedOwnerValue_1, resolvedFieldName_1, thisLocal.manager.config);
                    var resolvedSettedValue_1 = thisLocal.actionResolveSettedValue(action, fieldEtc_1);
                    resolvedSettedValue_1 = thisLocal.restoreEntireStateCallbackTemplate(function () {
                        var wasCollectionAsyncronousModified = { value: true };
                        switch (action.actionType) {
                            case TapeActionType.CollectionAdd:
                                if (resolvedCollection_1 && resolvedCollection_1.iAmLazyRef) {
                                    //in theory this should be synchronous because it is already created
                                    resolvedOwnerValue_1[resolvedFieldName_1]
                                        .subscribeToModify(function (coll) {
                                        thisLocal.addOnCollection(coll, resolvedSettedValue_1);
                                        wasCollectionAsyncronousModified.value = false;
                                    });
                                }
                                else {
                                    thisLocal.addOnCollection(resolvedCollection_1, resolvedSettedValue_1);
                                    wasCollectionAsyncronousModified.value = false;
                                }
                                if (wasCollectionAsyncronousModified.value) {
                                    throw new Error('Invalid action. Collection was not loaded on current state: ' + _this.jsonStringfyWithMax(action));
                                }
                                break;
                            case TapeActionType.CollectionRemove:
                                if (resolvedCollection_1 && resolvedCollection_1.iAmLazyRef) {
                                    //in theory this should be synchronous because it is already created
                                    resolvedOwnerValue_1[resolvedFieldName_1]
                                        .subscribeToModify(function (coll) {
                                        thisLocal.restoreEntireStateCallbackTemplate(function () {
                                            thisLocal.removeFromCollection(coll, resolvedSettedValue_1);
                                        });
                                        wasCollectionAsyncronousModified.value = false;
                                    });
                                }
                                else {
                                    thisLocal.restoreEntireStateCallbackTemplate(function () {
                                        thisLocal.removeFromCollection(resolvedCollection_1, resolvedSettedValue_1);
                                    });
                                    wasCollectionAsyncronousModified.value = false;
                                }
                                if (wasCollectionAsyncronousModified.value) {
                                    throw new Error('Invalid action. Collection was not loaded on current state: ' + _this.jsonStringfyWithMax(action));
                                }
                                break;
                            case TapeActionType.SetField:
                                if (resolvedOwnerValue_1[resolvedFieldName_1] && resolvedOwnerValue_1[resolvedFieldName_1].iAmLazyRef) {
                                    var lr = resolvedOwnerValue_1[resolvedFieldName_1];
                                    if (fieldEtc_1.propertyOptions.lazyDirectRawWrite) {
                                        lr.respObs = resolvedSettedValue_1;
                                    }
                                    else {
                                        lr.setLazyObjNoNext(resolvedSettedValue_1);
                                    }
                                }
                                else {
                                    thisLocal.restoreEntireStateCallbackTemplate(function () {
                                        resolvedOwnerValue_1[resolvedFieldName_1] = resolvedSettedValue_1;
                                    });
                                }
                                break;
                            case TapeActionType.Delete:
                                //nothing
                                break;
                            case TapeActionType.Save:
                                //nothing
                                break;
                            default:
                                throw new Error('This should not happen');
                        }
                    });
                }
            };
            for (var _a = 0, _b = tape.actions; _a < _b.length; _a++) {
                var action = _b[_a];
                _loop_2(action);
            }
        }
    };
    RecorderSessionDefault.prototype.resolveMetadatas = function (options) {
        var thisLocal = this;
        var valueOrliteral = options.object || options.literalObject || {};
        var refererObjectOrLiteral = options.refererObject || options.refererLiteralObject || {};
        var refererObjMd = { $iAmPlayerMetadatas$: true };
        var objectMd = { $iAmPlayerMetadatas$: true };
        var playerObjectIdMd = { $iAmPlayerMetadatas$: true };
        var refererObjMdFound = false;
        var objectMdFound = false;
        var playerObjectIdMdFound = false;
        if (LodashLike.has(valueOrliteral, this.manager.config.playerMetadatasName)) {
            objectMdFound = true;
            objectMd = LodashLike.get(valueOrliteral, this.manager.config.playerMetadatasName);
        }
        if (LodashLike.has(refererObjectOrLiteral, this.manager.config.playerMetadatasName)) {
            refererObjMdFound = true;
            refererObjMd = LodashLike.get(refererObjectOrLiteral, this.manager.config.playerMetadatasName);
        }
        //we are processing the metadata
        if (options.key === this.manager.config.playerMetadatasName
            && valueOrliteral.$iAmPlayerMetadatas$
            && LodashLike.has(valueOrliteral.$playerObjectId$, this.manager.config.playerMetadatasName)) {
            if (LodashLike.has(valueOrliteral.$playerObjectId$, this.manager.config.playerMetadatasName)) {
                playerObjectIdMdFound = true;
                playerObjectIdMd = LodashLike.get(valueOrliteral.$playerObjectId$, this.manager.config.playerMetadatasName);
            }
        }
        if (options.refMap) {
            if (refererObjMd.$id$ && refererObjMd.$isLazyUninitialized$ && !options.refMap.has(refererObjMd.$id$)) {
                var dummySignatureInstance = {};
                dummySignatureInstance[thisLocal.manager.config.playerMetadatasName] = refererObjMd;
                options.refMap.set(refererObjMd.$id$, dummySignatureInstance);
            }
            else if (refererObjMd.$idRef$) {
                var trackedInstance = options.refMap.get(refererObjMd.$idRef$);
                var trackedInstanceMd = LodashLike.get(trackedInstance, this.manager.config.playerMetadatasName);
                if (!trackedInstanceMd.$iAmPlayerMetadatas$) {
                    throw new Error('There is something wrong with metadatas on refered object by $idRef$:\n' + refererObjMd.$idRef$);
                }
                //here sign ref to another isLazyUninitialized metadata, this shrink the json.
                if (trackedInstanceMd.$isLazyUninitialized$) {
                    refererObjMd = trackedInstanceMd;
                }
            }
            if (objectMd.$id$ && objectMd.$isLazyUninitialized$ && !options.refMap.has(objectMd.$id$)) {
                var dummySignatureInstance = {};
                dummySignatureInstance[thisLocal.manager.config.playerMetadatasName] = objectMd;
                options.refMap.set(objectMd.$id$, dummySignatureInstance);
            }
            else if (objectMd.$idRef$) {
                var trackedInstance = options.refMap.get(objectMd.$idRef$);
                var trackedInstanceMd = LodashLike.get(trackedInstance, this.manager.config.playerMetadatasName);
                if (!trackedInstanceMd.$iAmPlayerMetadatas$) {
                    throw new Error('There is something wrong with refered object by $idRef$:\n' +
                        objectMd.$idRef$);
                }
                //here sign ref to another isLazyUninitialized metadata, this shrink the json.
                if (trackedInstanceMd.$isLazyUninitialized$) {
                    objectMd = trackedInstanceMd;
                }
            }
            if (playerObjectIdMd.$id$ && playerObjectIdMd.$isLazyUninitialized$ && !options.refMap.has(playerObjectIdMd.$id$)) {
                var dummySignatureInstance = {};
                dummySignatureInstance[thisLocal.manager.config.playerMetadatasName] = playerObjectIdMd;
                options.refMap.set(playerObjectIdMd.$id$, dummySignatureInstance);
            }
            else if (playerObjectIdMd.$idRef$) {
                var trackedInstance = options.refMap.get(playerObjectIdMd.$idRef$);
                var trackedInstanceMd = LodashLike.get(trackedInstance, this.manager.config.playerMetadatasName);
                if (!trackedInstanceMd.$iAmPlayerMetadatas$) {
                    throw new Error('There is something wrong with refered object by $idRef$:\n' +
                        playerObjectIdMd.$idRef$);
                }
                //here sign ref to another isLazyUninitialized metadata, this shrink the json.
                if (trackedInstanceMd.$isLazyUninitialized$) {
                    playerObjectIdMd = trackedInstanceMd;
                }
            }
        }
        return {
            refererObjMd: refererObjMd,
            objectMd: objectMd,
            playerObjectIdMd: playerObjectIdMd,
            refererObjMdFound: refererObjMdFound,
            objectMdFound: objectMdFound,
            playerObjectIdMdFound: playerObjectIdMdFound
        };
    };
    RecorderSessionDefault.prototype.createLiteralRefForEntity = function (realEntity) {
        if (!realEntity) {
            throw new Error('realEntity can not be null');
        }
        var allMD = this.resolveMetadatas({ object: realEntity });
        var entityRefReturn;
        var bMd = allMD.objectMd;
        if (bMd.$signature$) {
            entityRefReturn = {
                signatureStr: bMd.$signature$,
                iAmAnEntityRef: true
            };
        }
        else if (LodashLike.has(realEntity, this.manager.config.creationIdName)) {
            entityRefReturn = {
                creationId: LodashLike.get(realEntity, this.manager.config.creationIdName),
                iAmAnEntityRef: true
            };
        }
        else {
            throw new Error('Invalid operation. Not managed entity. Entity: \'' + realEntity.constructor + '\'');
        }
        return entityRefReturn;
    };
    RecorderSessionDefault.prototype.getEntityInstanceFromLiteralRef = function (literalRef) {
        var entityRef = literalRef;
        if (entityRef.iAmAnEntityRef && entityRef.signatureStr) {
            return this._objectsBySignature.get(entityRef.signatureStr);
        }
        else if (entityRef.iAmAnEntityRef && entityRef.creationId) {
            return this._objectsByCreationId.get(entityRef.creationId);
        }
        else {
            throw new Error('Invalid operation. Not managed entity. literalRef: \'' + literalRef + '\'');
        }
    };
    Object.defineProperty(RecorderSessionDefault.prototype, "manager", {
        /**
         * Getter manager
         * @return {RecorderManager}
         */
        get: function () {
            return this._manager;
        },
        /**
         * Setter manager
         * @param {RecorderManager} value
         */
        set: function (value) {
            var thisLocal = this;
            if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
                thisLocal.consoleLike.group('RecorderSessionDefault.manager set');
                thisLocal.consoleLike.debug(value);
                thisLocal.consoleLike.groupEnd();
            }
            this._manager = value;
        },
        enumerable: true,
        configurable: true
    });
    RecorderSessionDefault.prototype.processPlayerSnapshot = function (entityType, playerSnapshot) {
        var thisLocal = this;
        var result;
        thisLocal.validatePlayerSideLiteralObject(playerSnapshot);
        if (!playerSnapshot.wrappedSnapshot) {
            throw new Error('playerSnapshot.result exists: ' + this.jsonStringfyWithMax(playerSnapshot));
        }
        var playerTypeOptions = Reflect.getMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_TYPE, entityType);
        if (!playerTypeOptions) {
            throw new Error('the classe \'' + entityType + ' is not using the decorator \'RecorderDecorators.playerType\'');
        }
        var allMD = this.resolveMetadatas({ literalObject: playerSnapshot.wrappedSnapshot });
        var bMd = allMD.objectMd;
        if (!this.isOnRestoreEntireStateFromLiteral()) {
            if (!bMd.$isComponent$) {
                this.storeOriginalLiteralEntry({
                    method: 'processResultEntity',
                    reflectFunctionMetadataTypeKey: RecorderDecoratorsInternal.mountContructorByPlayerTypeMetadataKey(playerTypeOptions, entityType),
                    playerSnapshot: playerSnapshot
                });
            }
        }
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
            thisLocal.consoleLike.group('RecorderSessionDefault.processResultEntity<L>()');
            thisLocal.consoleLike.debug(entityType);
            thisLocal.consoleLike.debug(playerSnapshot);
            thisLocal.consoleLike.groupEnd();
        }
        var refMap = new Map();
        result = this.processResultEntityPriv(entityType, playerSnapshot.wrappedSnapshot, refMap);
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
            thisLocal.consoleLike.group('RecorderSessionDefault.processResultEntity<L>(). result:');
            thisLocal.consoleLike.debug(result);
            thisLocal.consoleLike.groupEnd();
        }
        return result;
    };
    RecorderSessionDefault.prototype.processPlayerSnapshotArray = function (entityType, playerSnapshot) {
        var thisLocal = this;
        var resultArr = [];
        if (!playerSnapshot.wrappedSnapshot) {
            throw new Error('playerSnapshot.result existe' + this.jsonStringfyWithMax(playerSnapshot));
        }
        var playerTypeOptions = Reflect.getMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_TYPE, entityType);
        if (!playerTypeOptions) {
            throw new Error('the classe \'' + entityType + ' is not using the decorator \'RecorderDecorators.playerType\'');
        }
        if (!this.isOnRestoreEntireStateFromLiteral()) {
            this.storeOriginalLiteralEntry({
                method: 'processResultEntityArray',
                reflectFunctionMetadataTypeKey: RecorderDecoratorsInternal.mountContructorByPlayerTypeMetadataKey(playerTypeOptions, entityType),
                playerSnapshot: playerSnapshot
            });
        }
        var refMap = new Map();
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
            thisLocal.consoleLike.group('RecorderSessionDefault.processResultEntityArray<L>()');
            thisLocal.consoleLike.debug(entityType);
            thisLocal.consoleLike.debug(playerSnapshot);
            thisLocal.consoleLike.groupEnd();
        }
        for (var index = 0; index < playerSnapshot.wrappedSnapshot.length; index++) {
            var resultElement = playerSnapshot.wrappedSnapshot[index];
            resultArr.push(this.processResultEntityPriv(entityType, resultElement, refMap));
        }
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
            thisLocal.consoleLike.group('RecorderSessionDefault.processResultEntityArray<L>(). wrappedSnapshot:');
            thisLocal.consoleLike.debug(playerSnapshot);
            thisLocal.consoleLike.groupEnd();
        }
        return resultArr;
    };
    RecorderSessionDefault.prototype.newEntityInstanceWithCreationId = function (entityType, creationId) {
        var thisLocal = this;
        if (!this.isOnRestoreEntireStateFromLiteral() && !this.isRecording()) {
            throw new Error('Invalid operation. It is not recording. is this Error correct?!');
        }
        this.validatingMetaFieldsExistence(entityType);
        var entityObj = new entityType();
        LodashLike.set(entityObj, RecorderConstants.ENTITY_SESION_PROPERTY_NAME, this);
        var realKeys = Object.keys(Object.getPrototypeOf(entityObj));
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
            thisLocal.consoleLike.debug('entityType: ' + entityType.name);
        }
        for (var index = 0; index < realKeys.length; index++) {
            var keyItem = realKeys[index];
            var fieldEtc = RecorderManagerDefault.resolveFieldProcessorPropOptsEtc(thisLocal._fielEtcCacheMap, entityObj, keyItem, thisLocal.manager.config);
            //let prpGenType: GenericNode = GenericTokenizer.resolveNode(entityObj, keyItem);
            if (!fieldEtc.prpGenType) {
                if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLike.debug('GenericNode not found for property key \'' + keyItem + '\' of ' + entityType.name);
                }
            }
            else if (fieldEtc.lazyRefMarkerType) {
                if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLike.debug('GenericNode found but it is not a LazyRef. Property key \'' + keyItem + '\' of ' + entityType.name);
                }
            }
            else {
                if (fieldEtc.lazyLoadedObjType) {
                    if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                        thisLocal.consoleLike.debug('GenericNode found and it is a LazyRef, fieldEtc.lazyLoadedObjType: ' + fieldEtc.lazyLoadedObjType.name + ' . Property key \'' + keyItem + '\' of ' + entityType.name);
                    }
                    var allMD = this.resolveMetadatas({ refererObject: entityObj, key: keyItem });
                    if (fieldEtc.lazyRefMarkerType === LazyRefOTMMarker) {
                        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                            thisLocal.consoleLike.debug('GenericNode found, it is a LazyRef, and it is a Collection, fieldEtc.otmCollectionType: ' + fieldEtc.otmCollectionType.name + ' . Property key \'' + keyItem + '\' of ' + entityType.name);
                        }
                        var lazyRefSet = new LazyRefOTMDefault(thisLocal);
                        lazyRefSet.setLazyObjOnLazyLoadingNoNext(this.createCollection(fieldEtc.otmCollectionType, entityObj, keyItem));
                        lazyRefSet.instanceId = this.nextMultiPurposeInstanceId();
                        lazyRefSet.refererObj = entityObj;
                        lazyRefSet.refererKey = keyItem;
                        lazyRefSet.session = this;
                        lazyRefSet.bMdLazyLoadedObj = allMD.objectMd;
                        lazyRefSet.bMdRefererObj = allMD.refererObjMd;
                        lazyRefSet.pbMdRefererPlayerObjectId = allMD.playerObjectIdMd;
                        LodashLike.set(entityObj, keyItem, lazyRefSet);
                    }
                    else if (fieldEtc.lazyRefMarkerType === LazyRefMTOMarker) {
                        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                            thisLocal.consoleLike.debug('GenericNode found, it is a LazyRefMTO, and it is not a Collection, fieldEtc.lazyLoadedObjType: ' + fieldEtc.lazyLoadedObjType.name + ' . Property key \'' + keyItem + '\' of ' + entityType.name);
                        }
                        var lazyRef = new LazyRefMTODefault(thisLocal);
                        lazyRef.instanceId = this.nextMultiPurposeInstanceId();
                        lazyRef.refererObj = entityObj;
                        lazyRef.refererKey = keyItem;
                        lazyRef.session = this;
                        lazyRef.bMdLazyLoadedObj = allMD.objectMd;
                        lazyRef.bMdRefererObj = allMD.refererObjMd;
                        lazyRef.pbMdRefererPlayerObjectId = allMD.playerObjectIdMd;
                        LodashLike.set(entityObj, keyItem, lazyRef);
                    }
                    else if (fieldEtc.lazyRefMarkerType === LazyRefMTOMarker) {
                        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                            thisLocal.consoleLike.debug('GenericNode found, it is a LazyRefMTO, and it is not a Collection, fieldEtc.lazyLoadedObjType: ' + fieldEtc.lazyLoadedObjType.name + ' . Property key \'' + keyItem + '\' of ' + entityType.name);
                        }
                        var lazyRef = new LazyRefPrpDefault(thisLocal);
                        lazyRef.instanceId = this.nextMultiPurposeInstanceId();
                        lazyRef.refererObj = entityObj;
                        lazyRef.refererKey = keyItem;
                        lazyRef.session = this;
                        lazyRef.bMdLazyLoadedObj = allMD.objectMd;
                        lazyRef.bMdRefererObj = allMD.refererObjMd;
                        lazyRef.pbMdRefererPlayerObjectId = allMD.playerObjectIdMd;
                        LodashLike.set(entityObj, keyItem, lazyRef);
                    }
                    else {
                        throw new Error('Property \'' + keyItem + ' of \'' + entityObj.constructor + '\'. LazyRef not properly defined on Reflect');
                    }
                }
                else {
                    throw new Error('Property \'' + keyItem + ' of \'' + entityObj.constructor + '\'. LazyRef not properly defined on Reflect');
                }
            }
        }
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
            thisLocal.consoleLike.debug('isRecording, ');
        }
        this._objectsByCreationId.set(creationId, entityObj);
        var playerTypeOptions = Reflect.getMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_TYPE, entityType);
        if (!playerTypeOptions) {
            throw new Error('the classe \'' + entityType + ' is not using the decorator \'RecorderDecorators.playerType\'');
        }
        if (!this.isOnRestoreEntireStateFromLiteral()) {
            this.storeOriginalLiteralEntry({
                method: 'newEntityInstance',
                reflectFunctionMetadataTypeKey: RecorderDecoratorsInternal.mountContructorByPlayerTypeMetadataKey(playerTypeOptions, entityType),
                ref: {
                    creationId: creationId,
                    iAmAnEntityRef: true
                }
            });
        }
        LodashLike.set(entityObj, this.manager.config.creationIdName, creationId);
        LodashLike.set(entityObj, RecorderConstants.ENTITY_SESION_PROPERTY_NAME, this);
        if (!this.isOnRestoreEntireStateFromLiteral()) {
            //recording tape
            var action = new TapeActionDefault();
            action.fieldName = null;
            action.actionType = TapeActionType.Create;
            var playerTypeOptions_1 = Reflect.getMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_TYPE, entityType);
            if (!playerTypeOptions_1) {
                throw new Error('the classe \'' + entityType + ' is not using the decorator \'RecorderDecorators.playerType\'');
            }
            action.ownerPlayerType = playerTypeOptions_1.playerType;
            action.ownerCreationId = this._nextCreationId;
            this.addTapeAction(action);
        }
        return entityObj;
    };
    RecorderSessionDefault.prototype.newEntityInstance = function (entityType) {
        if (!this.isRecording()) {
            throw new Error('Invalid operation. It is not recording.');
        }
        var newEntityInstanceWithCreationId = this.newEntityInstanceWithCreationId(entityType, this._nextCreationId);
        this._nextCreationId++;
        return newEntityInstanceWithCreationId;
    };
    RecorderSessionDefault.prototype.startRecording = function () {
        var thisLocal = this;
        if (this.isRecording()) {
            throw new Error('I am already recording!');
        }
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
            thisLocal.consoleLike.debug('reseting this._currentTape, this._objectsCreationId and this._nextCreationId');
        }
        this._currentTape = new TapeDefault();
        this._nextCreationId = 1;
    };
    RecorderSessionDefault.prototype.stopRecording = function () {
        var thisLocal = this;
        if (this.isRecording()) {
            if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
                thisLocal.consoleLike.debug('updating this.lastTape and resetting this.currentTape');
            }
            this._latestTape.push(this._currentTape);
            this._currentTape = null;
        }
        else {
            throw new Error('I am not recording!');
        }
    };
    RecorderSessionDefault.prototype.recordSave = function (entity) {
        var thisLocal = this;
        if (!entity) {
            throw new Error('entity can not be null');
        }
        if (!this.isRecording()) {
            throw new Error('Invalid operation. It is not recording. entity: \'' + entity.constructor.name + '\'. Is this Error correct?!');
        }
        var session = LodashLike.get(entity, RecorderConstants.ENTITY_SESION_PROPERTY_NAME);
        if (!session) {
            throw new Error('Invalid operation. \'' + entity.constructor.name + '\' not managed. \'' + RecorderConstants.ENTITY_SESION_PROPERTY_NAME + '\' estah null');
        }
        else if (session !== this) {
            throw new Error('Invalid operation. \'' + entity.constructor.name + '\' managed by another session.');
        }
        var allMD = this.resolveMetadatas({ object: entity });
        var bMd = allMD.objectMd;
        //recording tape
        var action = new TapeActionDefault();
        action.actionType = TapeActionType.Save;
        if (bMd.$signature$) {
            throw new Error('Invalid operation. \'' + entity.constructor + '\' has a signature, that is, it has persisted');
        }
        else if (LodashLike.has(entity, this.manager.config.creationIdName)) {
            action.ownerCreationRefId = LodashLike.get(entity, this.manager.config.creationIdName);
        }
        else {
            throw new Error('Invalid operation. Not managed entity. Entity: \'' + entity.constructor + '\'');
        }
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
            thisLocal.consoleLike.group('action: ');
            thisLocal.consoleLike.debug(action);
            thisLocal.consoleLike.groupEnd();
        }
        this.addTapeAction(action);
    };
    RecorderSessionDefault.prototype.recordDelete = function (entity) {
        var thisLocal = this;
        if (!entity) {
            throw new Error('Entity nao pode ser nula');
        }
        if (!this.isRecording()) {
            throw new Error('Invalid operation. It is not recording. entity: \'' + entity.constructor.name + '\'. Is this Error correct?!');
        }
        var session = LodashLike.get(entity, RecorderConstants.ENTITY_SESION_PROPERTY_NAME);
        if (!session) {
            throw new Error('Invalid operation. \'' + entity.constructor + '\' not managed. \'' + RecorderConstants.ENTITY_SESION_PROPERTY_NAME + '\' estah null');
        }
        else if (session !== this) {
            throw new Error('Invalid operation. \'' + entity.constructor + '\' managed by another session.');
        }
        var allMD = this.resolveMetadatas({ object: entity });
        var bMd = allMD.objectMd;
        //recording tape
        var action = new TapeActionDefault();
        action.actionType = TapeActionType.Delete;
        if (bMd.$signature$) {
            action.ownerSignatureStr = bMd.$signature$;
        }
        else if (LodashLike.has(entity, this.manager.config.creationIdName)) {
            throw new Error('Invalid operation. \'' + entity.constructor + '\' has id of creation, that is, is not persisted.');
        }
        else {
            throw new Error('Invalid operation. Not managed entity. Entity: \'' + entity.constructor + '\'');
        }
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
            thisLocal.consoleLike.debug('action: ' + action);
        }
        this.addTapeAction(action);
    };
    RecorderSessionDefault.prototype.recordAtache = function (attach) {
        var name = this.manager.config.attachPrefix + this.nextMultiPurposeInstanceId();
        this._currentRecordedAtaches.set(name, attach);
        return name;
    };
    RecorderSessionDefault.prototype.storeOriginalLiteralEntry = function (originalValueEntry) {
        this._originalLiteralValueEntries.push(originalValueEntry);
    };
    RecorderSessionDefault.prototype.clear = function () {
        var thisLocal = this;
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
            thisLocal.consoleLike.debug('clearing: this.objectsBySignature, this.objectsCreationId, this.nextCreationId, this.currentTape, this.lastTape and this.objectsBySignature');
        }
        this._nextCreationId = null;
        this._currentTape = null;
        this._latestTape = null;
        this._nextMultiPurposeInstanceId = 1;
        this._objectsBySignature = new Map();
        this._objectsByCreationId = new Map();
        this._lazyrefsByEntityMap = new Map();
        this._fielEtcCacheMap = new Map();
        this._switchedOffNotificationEntitiesSet = new Set();
        this._originalLiteralValueEntries = [];
        this._latestTape = [];
        var clearCache$ = this.manager.config.cacheHandler.clearCache();
        // clearCache$ = clearCache$.pipe(this.addSubscribedObsRxOpr());
        return clearCache$.pipe(thisLocal.registerProvidedObservablesRxOpr(), share());
    };
    RecorderSessionDefault.prototype.getLastRecordedTape = function () {
        var thisLocal = this;
        return thisLocal._latestTape.length > 0 ? thisLocal._latestTape[thisLocal._latestTape.length - 1] : null;
    };
    RecorderSessionDefault.prototype.getLastRecordedStreams = function () {
        var thisLocal = this;
        var tape = this.getLastRecordedTape();
        var idAndStreamArr = [];
        if (tape && tape.actions) {
            for (var _i = 0, _a = tape.actions; _i < _a.length; _i++) {
                var actionItem = _a[_i];
                if (actionItem.attachRefId) {
                    var stream$ = thisLocal.manager.config.cacheHandler.getFromCache(actionItem.attachRefId)
                        .pipe(mapJustOnceRxOpr(function (streamValue) {
                        return streamValue;
                    }), share());
                    idAndStreamArr.push({
                        attachRefId: actionItem.attachRefId,
                        stream: stream$
                    });
                }
            }
        }
        var resultMap = new Map();
        for (var _b = 0, idAndStreamArr_1 = idAndStreamArr; _b < idAndStreamArr_1.length; _b++) {
            var idAndStreamItem = idAndStreamArr_1[_b];
            resultMap.set(idAndStreamItem.attachRefId, idAndStreamItem.stream);
        }
        return resultMap;
    };
    RecorderSessionDefault.prototype.getLastRecordedTapeAndStreams = function () {
        var thisLocal = this;
        var tape = this.getLastRecordedTape();
        var streamsMap = thisLocal.getLastRecordedStreams();
        return {
            tape: tape,
            streams: streamsMap
        };
    };
    RecorderSessionDefault.prototype.getLastRecordedTapeAsLiteralAndStreams = function () {
        var thisLocal = this;
        var tapeLiteral = this.getLastRecordedTapeAsLiteral();
        var streamsMap = thisLocal.getLastRecordedStreams();
        return {
            tapeLiteral: tapeLiteral,
            streams: streamsMap
        };
    };
    RecorderSessionDefault.prototype.addTapeAction = function (action) {
        var thisLocal = this;
        if (!this.isRecording()) {
            throw new Error('The recording is not started!');
        }
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
            thisLocal.consoleLike.group('addTapeAction');
            thisLocal.consoleLike.debug(action);
            thisLocal.consoleLike.groupEnd();
        }
        this._currentTape.actions.push(action);
    };
    RecorderSessionDefault.prototype.isRecording = function () {
        return (this._currentTape != null);
    };
    RecorderSessionDefault.prototype.getLastRecordedTapeAsLiteral = function () {
        var thisLocal = this;
        var tape = this.getLastRecordedTape();
        var playbackAsLiteral = thisLocal.getPlaybackAsLiteral(tape);
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
            thisLocal.consoleLike.group('getLastRecordedTapeAsLiteral');
            thisLocal.consoleLike.debug(playbackAsLiteral);
            thisLocal.consoleLike.groupEnd();
        }
        return playbackAsLiteral;
    };
    RecorderSessionDefault.prototype.getLastRecordedAtaches = function () {
        return new Map(this._currentRecordedAtaches);
    };
    RecorderSessionDefault.prototype.getPlaybackAsLiteral = function (tape) {
        var thisLocal = this;
        var literalReturn = JSONHelper.convertToLiteralObject(tape, true);
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
            thisLocal.consoleLike.group('getPlaybackAsLiteral');
            thisLocal.consoleLike.debug(literalReturn);
            thisLocal.consoleLike.groupEnd();
        }
        return literalReturn;
    };
    RecorderSessionDefault.prototype.getTapeFromLiteral = function (tapeLiteral) {
        var thisLocal = this;
        var playBackReturn = new TapeDefault();
        playBackReturn.actions = [];
        for (var _i = 0, _a = tapeLiteral.actions; _i < _a.length; _i++) {
            var actionLiteral = _a[_i];
            var action = new TapeActionDefault();
            action = LodashLike.mergeWith(action, actionLiteral, function (value, srcValue) {
                return srcValue;
            });
            playBackReturn.actions.push(action);
        }
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Debug)) {
            thisLocal.consoleLike.group('getPlaybackFromLiteral');
            thisLocal.consoleLike.debug(playBackReturn);
            thisLocal.consoleLike.groupEnd();
        }
        return playBackReturn;
    };
    RecorderSessionDefault.prototype.getCachedBySignature = function (signatureStr) {
        if (this._objectsBySignature.get(signatureStr)) {
            return this._objectsBySignature.get(signatureStr);
        }
        else {
            return null;
        }
    };
    RecorderSessionDefault.prototype.validatingMetaFieldsExistence = function (entityType) {
        var camposControleArr = [
            this.manager.config.creationIdName,
            this.manager.config.playerMetadatasName,
            RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME,
            RecorderConstants.ENTITY_SESION_PROPERTY_NAME
        ];
        for (var index = 0; index < camposControleArr.length; index++) {
            var internalKeyItem = camposControleArr[index];
            if (Object.keys(entityType.prototype).lastIndexOf(internalKeyItem.toString()) >= 0) {
                throw new Error('The Entity ' + entityType.name + ' already has the property \'' + internalKeyItem.toString() + '\' offined!');
            }
        }
    };
    RecorderSessionDefault.prototype.validatePlayerSideLiteralObject = function (literalObject) {
        var allMD = this.resolveMetadatas({ literalObject: literalObject });
        if (!allMD.objectMd) {
            throw new Error('Invalid player side literal object: \n' + this.jsonStringfyWithMax(literalObject));
        }
    };
    RecorderSessionDefault.prototype.validatePlayerSideResponseLike = function (responseLike) {
        var isResponseBodyStream = responseLike.body && responseLike.body.pipe;
        if (!isResponseBodyStream) {
            if (!responseLike || !responseLike.body || !responseLike.body.wrappedSnapshot) {
                throw new Error('Invalid player side responseLike body object: \n' + this.jsonStringfyWithMax(responseLike.body));
            }
        }
    };
    RecorderSessionDefault.prototype.processWrappedSnapshotFieldArrayInternal = function (entityType, lazyLoadedColl, wrappedSnapshotField) {
        var thisLocal = this;
        var refMap = new Map();
        var realItemObsArr = [];
        if (!Array.isArray(wrappedSnapshotField)) {
            throw new Error('wrappedSnapshotField is not an Array:\n' + this.jsonStringfyWithMax(wrappedSnapshotField));
        }
        for (var _i = 0, wrappedSnapshotField_1 = wrappedSnapshotField; _i < wrappedSnapshotField_1.length; _i++) {
            var literalItem = wrappedSnapshotField_1[_i];
            var realItem = this.processResultEntityPriv(entityType, literalItem, refMap);
            realItemObsArr.push(realItem);
        }
        thisLocal.lazyLoadTemplateCallback(lazyLoadedColl, function () {
            for (var _i = 0, realItemObsArr_1 = realItemObsArr; _i < realItemObsArr_1.length; _i++) {
                var realItem = realItemObsArr_1[_i];
                thisLocal.addOnCollection(lazyLoadedColl, realItem);
            }
        });
    };
    RecorderSessionDefault.prototype.processWrappedSnapshotFieldInternal = function (entityType, snapshotField) {
        var refMap = new Map();
        var result = this.processResultEntityPriv(entityType, snapshotField, refMap);
        return result;
    };
    RecorderSessionDefault.prototype.processResultEntityPriv = function (entityType, wrappedSnapshotField, refMap) {
        var _this = this;
        var thisLocal = this;
        if (!wrappedSnapshotField) {
            throw new Error('snapshotField can not be null');
        }
        this.validatePlayerSideLiteralObject(wrappedSnapshotField);
        var allMD = this.resolveMetadatas({ literalObject: wrappedSnapshotField, refMap: refMap });
        var bMd = allMD.objectMd;
        var entityValue = this._objectsBySignature.get(bMd.$signature$);
        //resultEntityAlreadyProcessed.allMD = allMD;
        if (!LodashLike.isNil(entityValue)) {
            if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                thisLocal.consoleLike.debug('entity is already processed on this session. Found by signature: ' + bMd.$signature$);
            }
            if (!bMd.$id$) {
                throw new Error('entity is already processed on this session. But thre is no metadatas.$id$. signature: ' + bMd.$signature$);
            }
            refMap.set(bMd.$id$, entityValue);
            LodashLike.set(entityValue, RecorderConstants.ENTITY_EXISTS_BY_SIGN_FROM_PREVIOUS_PROCESSING, true);
        }
        else {
            if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                thisLocal.consoleLike.debug('entity was not processed yet on this session. Not found by signature: ' + bMd.$signature$);
            }
            if (bMd.$idRef$) {
                entityValue = refMap.get(bMd.$idRef$);
                if (LodashLike.isNil(entityValue)) {
                    throw new Error('entity not foun for idRef: ' + bMd.$idRef$);
                }
            }
        }
        if (LodashLike.isNil(entityValue)) {
            entityValue = new entityType();
            if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                thisLocal.consoleLike.debug('entity was not processed yet on this session.' +
                    ' Creating new instance for: ' + entityType.name);
            }
        }
        //const breackPointFlag = { fooid: ''};
        if (!wrappedSnapshotField) {
            throw new Error('snapshotField can not be null');
        }
        // if (LodashLike.get(entityValue, RecorderConstants.ENTITY_EXISTS_BY_SIGN_FROM_PREVIOUS_PROCESSING)) {
        this.validatingMetaFieldsExistence(entityType);
        // entityValue = resultEntityAlreadyProcessed.entityValue;
        LodashLike.set(entityValue, RecorderConstants.ENTITY_SESION_PROPERTY_NAME, this);
        this.removeNonUsedKeysFromLiteral(entityValue, wrappedSnapshotField);
        if (bMd.$id$) {
            refMap.set(bMd.$id$, entityValue);
        }
        else {
            if (!bMd.$idRef$) {
                throw new Error('This should not happen 1');
            }
        }
        if (!bMd.$idRef$) {
            //literal source values with $idRef$ no need deep processing!
            entityValue = this.lazyLoadTemplateCallback(entityValue, function () {
                _this.tryCacheInstanceBySignature({
                    realInstance: entityValue,
                    playerSnapshot: { wrappedSnapshot: wrappedSnapshotField }
                });
                return LodashLike.asyncMergeWith(entityValue, wrappedSnapshotField, {
                    customizer: _this.mergeWithCustomizerPropertyReplection(refMap),
                    noObjects: new Set([Date]),
                    considerObjectProperties: true,
                    ignorePropeties: [
                        /^_.*/,
                        RecorderConstants.ENTITY_CONTROL_PROPS_PATTERN
                    ]
                });
            });
        }
        return entityValue;
    };
    RecorderSessionDefault.prototype.createLoadedLazyRef = function (fieldEtc, literalLazyObj, refMap, refererObj, refererKey) {
        var thisLocal = this;
        var lr = this.createApropriatedLazyRef(fieldEtc.prpGenType, literalLazyObj, refererObj, refererKey, refMap);
        var allMD = thisLocal.resolveMetadatas({ literalObject: literalLazyObj, refererObject: refererObj, key: refererKey, refMap: refMap });
        this.trySetPlayerObjectIdentifier(lr, fieldEtc.prpGenType, literalLazyObj, refMap);
        this.tryGetFromObjectsBySignature(lr, literalLazyObj, refMap);
        var isValueByFieldProcessor = { value: false };
        if (allMD.objectMd.$iAmPlayerMetadatas$) {
            if (allMD.objectMd.$idRef$) {
                var referedInstance = refMap.get(allMD.objectMd.$idRef$);
                if (!literalLazyObj) {
                    throw new Error('literalLazyObj.$iAmPlayerMetadatas$ and $idRef$ not found: \'' + refererKey + '\' on ' + refererObj.constructor);
                }
                lr.setLazyObjOnLazyLoadingNoNext(referedInstance);
                //asyncCombineObsArr.push(setLazyObjOnLazyLoadingNoNext$);
            }
            else {
                //nothing
            }
        }
        else {
            //nothing
        }
        ;
        if (LodashLike.isNil(literalLazyObj)) {
            //LodashLike.isNil(srcValue) means LazyRef object instance is null.
            if (!fieldEtc.propertyOptions.lazyDirectRawWrite) {
                lr.setLazyObjOnLazyLoadingNoNext(null);
            }
            else {
                lr.setRealResponseDoneDirectRawWrite(true);
                lr.respObs = of({ body: null });
            }
        }
        else {
            if (lr.lazyLoadedObj) {
                if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLike.group('LazyRef.lazyLoadedObj is already setted: ');
                    thisLocal.consoleLike.debug(lr.lazyLoadedObj);
                    thisLocal.consoleLike.groupEnd();
                }
                //nothing
            }
            else {
                if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLike.debug('LazyRef.lazyLoadedObj is not setted yet');
                }
                //let lazyLoadedObjType: TypeLike<any> = null;
                if (fieldEtc.lazyRefMarkerType === LazyRefOTMMarker) {
                    if (!fieldEtc.otmCollectionType) {
                        throw new Error('Property \'' + refererKey + ' of \'' + refererObj.constructor.name + '\'. . LazyRefOTM not properly defined on Reflect. Collection Type no denined!');
                    }
                    var lazyCollection_1 = this.createCollection(fieldEtc.otmCollectionType, refererObj, refererKey);
                    thisLocal.lazyLoadTemplateCallback(lazyCollection_1, function () {
                        lr.setLazyObjOnLazyLoadingNoNext(lazyCollection_1);
                        var processResultEntityPrivArr = [];
                        for (var _i = 0, literalLazyObj_1 = literalLazyObj; _i < literalLazyObj_1.length; _i++) {
                            var literalItem = literalLazyObj_1[_i];
                            var processResultEntityPriv = thisLocal.processResultEntityPriv(fieldEtc.lazyLoadedObjType, literalItem, refMap);
                            processResultEntityPrivArr.push(processResultEntityPriv);
                        }
                        for (var _a = 0, processResultEntityPrivArr_1 = processResultEntityPrivArr; _a < processResultEntityPrivArr_1.length; _a++) {
                            var entityItem = processResultEntityPrivArr_1[_a];
                            thisLocal.addOnCollection(lazyCollection_1, entityItem);
                        }
                    });
                }
                else {
                    var fieldEtc_2 = RecorderManagerDefault.resolveFieldProcessorPropOptsEtc(this.fielEtcCacheMap, refererObj, refererKey, this.manager.config);
                    if (fieldEtc_2.lazyRefMarkerType === LazyRefPrpMarker) {
                        if (fieldEtc_2.fieldProcessorCaller.callFromLiteralValue) {
                            isValueByFieldProcessor.value = true;
                            var fromLiteralValue = fieldEtc_2.fieldProcessorCaller.callFromLiteralValue(literalLazyObj, fieldEtc_2.fieldInfo);
                            // callFromLiteralValue$ = callFromLiteralValue$.pipe(this.addSubscribedObsRxOpr());
                            //here for debug purpose
                            refererKey === 'blobLazyA';
                            if (!fieldEtc_2.propertyOptions.lazyDirectRawWrite) {
                                lr.setLazyObjOnLazyLoadingNoNext(fromLiteralValue);
                            }
                            else {
                                lr.attachRefId = thisLocal.manager.config.cacheStoragePrefix + thisLocal.nextMultiPurposeInstanceId();
                                lr.setRealResponseDoneDirectRawWrite(true);
                                lr.respObs = fieldEtc_2.fieldProcessorCaller.callToDirectRaw(fromLiteralValue, fieldEtc_2.fieldInfo).pipe(flatMap(function (respDirRaw) {
                                    return thisLocal.manager.config.cacheHandler.putOnCache(lr.attachRefId, respDirRaw.body);
                                }), share(), flatMap(function () {
                                    return thisLocal.manager.config.cacheHandler.getFromCache(lr.attachRefId);
                                }), map(function (stream) {
                                    return { body: stream };
                                }));
                            }
                        }
                        else {
                            if (!fieldEtc_2.propertyOptions.lazyDirectRawWrite) {
                                lr.setLazyObjOnLazyLoadingNoNext(literalLazyObj);
                            }
                            else {
                                throw new Error('LazyRef is lazyDirectRawWrite and has no fromLiteralValue processor.\n' + lr.toString());
                            }
                        }
                    }
                    else if (!isValueByFieldProcessor.value) {
                        var resultEntityPriv = this.processResultEntityPriv(fieldEtc_2.lazyLoadedObjType, literalLazyObj, refMap);
                        lr.setLazyObjOnLazyLoadingNoNext(resultEntityPriv);
                    }
                }
            }
        }
        return lr;
    };
    RecorderSessionDefault.prototype.tryCacheInstanceBySignature = function (tryOptions) {
        if (!tryOptions) {
            throw new Error('tryOptions can not be null');
        }
        if (!tryOptions.playerSnapshot) {
            throw new Error('tryOptions.playerSnapshot can not be null');
        }
        var allMD = this.resolveMetadatas({ literalObject: tryOptions.playerSnapshot.wrappedSnapshot });
        var bMd = allMD.objectMd;
        if (bMd.$signature$) {
            this._objectsBySignature.set(bMd.$signature$, tryOptions.realInstance);
        }
        if (tryOptions.lazySignature) {
            this._objectsBySignature.set(tryOptions.lazySignature, tryOptions.realInstance);
        }
    };
    RecorderSessionDefault.prototype.createNotLoadedLazyRef = function (genericNode, literalLazyObj, refMap, refererObj, refererKey) {
        var thisLocal = this;
        //const asyncCombineObsArr: Observable<any>[] = [];
        var propertyOptions = Reflect.getMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_OBJECT_PROPERTY_OPTIONS, refererObj, refererKey);
        if (!propertyOptions) {
            throw new Error('@RecorderDecorators.property() not defined for ' + refererObj.constructor.name + '.' + refererKey);
        }
        var lr = this.createApropriatedLazyRef(genericNode, literalLazyObj, refererObj, refererKey, refMap);
        this.trySetPlayerObjectIdentifier(lr, genericNode, literalLazyObj, refMap);
        this.tryGetFromObjectsBySignature(lr, literalLazyObj, refMap);
        if (lr.lazyLoadedObj) {
            if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                thisLocal.consoleLike.group('LazyRef.lazyLoadedObj is already setted: ');
                thisLocal.consoleLike.debug(lr.lazyLoadedObj);
                thisLocal.consoleLike.groupEnd();
            }
        }
        else {
            var lazyInfo = {
                gNode: genericNode,
                propertyOptions: propertyOptions,
                literalLazyObj: literalLazyObj,
                ownerType: refererObj.constructor,
                lazyFieldType: genericNode.gParams[0],
                fieldName: refererKey
            };
            if (!propertyOptions.lazyDirectRawRead) {
                if (!lr.signatureStr) {
                    throw new Error('Signature not found\n' + lr.toString());
                }
                lr.respObs = this.manager.config.lazyObservableProvider.generateObservable(lr.signatureStr, lazyInfo);
            }
            else {
                if (!lr.signatureStr) {
                    throw new Error('Signature not found\n' + lr.toString());
                }
                lr.respObs = this.manager.config.lazyObservableProvider.generateObservableForDirectRaw(lr.signatureStr, lazyInfo);
            }
        }
        return lr;
    };
    RecorderSessionDefault.prototype.tryGetFromObjectsBySignature = function (lr, literalLazyObj, refMap) {
        var allMD = this.resolveMetadatas({ literalObject: literalLazyObj, refMap: refMap });
        //let allMD = this.resolveMetadatas({literalObject: literalLazyObj});
        var bMd = allMD.objectMd;
        var entityValue = null;
        if (bMd.$signature$) {
            lr.signatureStr = bMd.$signature$;
            entityValue = this._objectsBySignature.get(bMd.$signature$);
            if (entityValue) {
                lr.setLazyObjOnLazyLoadingNoNext(entityValue);
            }
        }
        else {
        }
    };
    RecorderSessionDefault.prototype.createApropriatedLazyRef = function (genericNode, literalLazyObj, refererObj, refererKey, refMap) {
        var allMD = this.resolveMetadatas({ literalObject: literalLazyObj, refererObject: refererObj, key: refererKey, refMap: refMap });
        var bMd = allMD.objectMd;
        var playerObjectIdLiteral = bMd.$playerObjectId$;
        var lazyRef = null;
        var fieldEtc = RecorderManagerDefault.resolveFieldProcessorPropOptsEtc(this._fielEtcCacheMap, refererObj, refererKey, this.manager.config);
        if (fieldEtc.lazyRefMarkerType === LazyRefMTOMarker) {
            lazyRef = new LazyRefMTODefault(this);
            lazyRef.pbMdRefererPlayerObjectId;
        }
        else if (fieldEtc.lazyRefMarkerType === LazyRefOTMMarker) {
            if (!fieldEtc.otmCollectionType) {
                throw new Error('Property \'' + refererKey + ' of \'' + refererObj.constructor.name + '\'. . LazyRefOTM not properly defined on Reflect. Collection Type no defined!');
            }
            lazyRef = new LazyRefOTMDefault(this);
        }
        else if (fieldEtc.lazyRefMarkerType === LazyRefPrpMarker) {
            lazyRef = new LazyRefPrpDefault(this);
        }
        else {
            throw new Error('Property \'' + refererKey + '\' of \'' + refererObj.constructor.name + '\'. LazyRef not properly defined on Reflect');
        }
        lazyRef.instanceId = this.nextMultiPurposeInstanceId();
        lazyRef.refererObj = refererObj;
        lazyRef.refererKey = refererKey;
        lazyRef.session = this;
        lazyRef.genericNode = genericNode;
        lazyRef.bMdLazyLoadedObj = allMD.objectMd;
        lazyRef.bMdRefererObj = allMD.refererObjMd;
        lazyRef.pbMdRefererPlayerObjectId = allMD.refererObjMd;
        return lazyRef;
    };
    RecorderSessionDefault.prototype.isLiteralObjMetadataKey = function (keyName) {
        if (this.metadaKeys == null) {
            this.metadaKeys = new Set()
                .add(this.manager.config.playerMetadatasName);
        }
        return this.metadaKeys.has(keyName);
    };
    RecorderSessionDefault.prototype.removeNonUsedKeysFromLiteral = function (realObj, literalObj) {
        var literalKeys = LodashLike.clone(LodashLike.keys(literalObj));
        var realKeys = Object.keys(Object.getPrototypeOf(realObj));
        for (var index = 0; index < literalKeys.length; index++) {
            var keyItem = literalKeys[index];
            if (!this.isLiteralObjMetadataKey(keyItem) && realKeys.indexOf(keyItem) < 0) {
                delete literalObj[keyItem];
            }
        }
    };
    RecorderSessionDefault.prototype.trySetPlayerObjectIdentifier = function (lr, genericNode, literalLazyObj, refMap) {
        var thisLocal = this;
        var allMD = this.resolveMetadatas({ literalObject: literalLazyObj, refMap: refMap });
        var bMd = allMD.objectMd;
        var playerObjectIdLiteralRef = { value: undefined };
        if (bMd.$idRef$) {
            var referedInstance = refMap.get(bMd.$idRef$);
            var referedInstanceMd = LodashLike.get(referedInstance, thisLocal.manager.config.playerMetadatasName);
            if (!referedInstanceMd.$iAmPlayerMetadatas$) {
                throw new Error('Where is the metadatas fot $idRef$:\n' + bMd.$idRef$);
            }
            playerObjectIdLiteralRef.value = referedInstanceMd.$playerObjectId$;
        }
        else {
            playerObjectIdLiteralRef.value = bMd.$playerObjectId$;
        }
        if (LodashLike.isObject(playerObjectIdLiteralRef.value, new Set([Date, Buffer]))) {
            var playerObjectIdType = null;
            if (genericNode.gParams[1] instanceof GenericNode) {
                playerObjectIdType = genericNode.gParams[1].gType;
            }
            else {
                playerObjectIdType = genericNode.gParams[1];
            }
            if (playerObjectIdType) {
                if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLike.debug('There is a playerObjectIdType on LazyRef. Is it many-to-one LazyRef?!. playerObjectIdType: ' + playerObjectIdType.name + ', genericNode:' + genericNode);
                }
                this.validatingMetaFieldsExistence(playerObjectIdType);
                lr.playerObjectId = this.processResultEntityPriv(playerObjectIdType, playerObjectIdLiteralRef.value, refMap);
            }
            else {
                if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLike.debug('Thre is no playerObjectIdType on LazyRef. Is it a collection?!. playerObjectIdType: ' + playerObjectIdType.name + ', genericNode:' + genericNode);
                }
            }
        }
        else if (playerObjectIdLiteralRef.value) {
            if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                thisLocal.consoleLike.debug('The player object Id is a simple type value: ' + playerObjectIdLiteralRef.value + '. genericNode:' + genericNode);
            }
            lr.playerObjectId = playerObjectIdLiteralRef.value;
        }
        else {
            if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                thisLocal.consoleLike.debug('The player object Id is null! Is it a collection?!: ' + playerObjectIdLiteralRef.value + '. genericNode:' + genericNode);
            }
        }
    };
    RecorderSessionDefault.prototype.createCollection = function (collType, refererObj, refererKey) {
        if (collType === Set) {
            return new SetCreator(this, refererObj, refererKey).createByProxy();
        }
        else {
            throw new Error('Collection not supported: ' + collType);
        }
    };
    RecorderSessionDefault.prototype.isCollection = function (typeTested) {
        return RecorderManagerDefault.isCollection(typeTested);
    };
    RecorderSessionDefault.prototype.addOnCollection = function (collection, element) {
        if (collection instanceof Array) {
            throw new Error('Collection not supported: ' + collection.prototype);
        }
        else if (collection instanceof Set) {
            collection.add(element);
        }
        else {
            throw new Error('Collection not supported: ' + collection.prototype);
        }
    };
    RecorderSessionDefault.prototype.removeFromCollection = function (collection, element) {
        if (collection instanceof Array) {
            throw new Error('Collection not supported: ' + collection.prototype);
        }
        else if (collection instanceof Set) {
            collection.delete(element);
        }
        else {
            throw new Error('Collection not supported: ' + collection.prototype);
        }
    };
    /**
     * Used exclusively in lazy load or processing literal values from server, it does not create tape actions.
     * @param refMap
     */
    RecorderSessionDefault.prototype.mergeWithCustomizerPropertyReplection = function (refMap) {
        var _this = this;
        var thisLocal = this;
        return function (value, srcValue, key, object, source) {
            var keepAllFlagsTemplateCallback = thisLocal.createKeepAllFlagsTemplateCallback(object);
            if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                thisLocal.consoleLikeMerge.group('mergeWithCustomizerPropertyReplection => function');
                thisLocal.consoleLikeMerge.debug(refMap);
                thisLocal.consoleLikeMerge.debug(value);
                thisLocal.consoleLikeMerge.debug(srcValue);
                thisLocal.consoleLikeMerge.debug(key);
                thisLocal.consoleLikeMerge.debug(object);
                thisLocal.consoleLikeMerge.debug(source);
                thisLocal.consoleLikeMerge.groupEnd();
            }
            var allMD = thisLocal.resolveMetadatas({ literalObject: srcValue, refererLiteralObject: source, key: key, refMap: refMap });
            var mdSource = allMD.refererObjMd;
            var mdSrcValue = allMD.objectMd;
            var mdPlayerObjectId = allMD.playerObjectIdMd;
            var mdSrcValueFound = allMD.objectMdFound;
            var fieldEtc = RecorderManagerDefault.resolveFieldProcessorPropOptsEtc(thisLocal.fielEtcCacheMap, object, key, thisLocal.manager.config);
            var isLazyRefField = (fieldEtc
                && fieldEtc.prpGenType
                && (fieldEtc.lazyRefMarkerType === LazyRefMTOMarker
                    || fieldEtc.lazyRefMarkerType === LazyRefOTMMarker
                    || fieldEtc.lazyRefMarkerType === LazyRefPrpMarker));
            if (mdPlayerObjectId.$isComponent$) {
                if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLikeMerge.group('mergeWithCustomizerPropertyReplection => function: bMdPlayerObjectId.isComponent. bMdSrcValue.$playerObjectId$:');
                    thisLocal.consoleLikeMerge.debug(mdSrcValue.$playerObjectId$);
                    thisLocal.consoleLikeMerge.groupEnd();
                }
                //fieldEtc.prpType = Reflect.getMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_OBJECT_ID_TYPE, object);
                if (!fieldEtc.objectIdPrpType) {
                    throw new Error('We are receiving mdSrcValue.$playerObjectId$ as Object and mdPlayerObjectId.$isComponent$, ' + object.constructor.name + ' does not define a property with @JsonPlayback.playerObjectId()');
                }
            }
            if (mdSrcValue.$isAssociative$ && fieldEtc.prpGenType
                && fieldEtc.lazyRefMarkerType !== LazyRefMTOMarker
                && fieldEtc.lazyRefMarkerType !== LazyRefOTMMarker
                && fieldEtc.lazyRefMarkerType !== LazyRefPrpMarker) {
                throw new Error('Key ' + object.constructor.name + '.' + key + ' is player side associative relation and is not LazyRef or not define GenericTokenizer');
            }
            if (mdSrcValue.$isComponent$
                && fieldEtc.prpGenType
                && (fieldEtc.lazyRefMarkerType === LazyRefMTODefault
                    || fieldEtc.lazyRefMarkerType === LazyRefOTMMarker
                    || fieldEtc.lazyRefMarkerType === LazyRefPrpMarker)) {
                throw new Error('Key ' + object.constructor.name + '.' + key + ' is player side component and is a LazyRef.');
            }
            var customizerResult = { needSet: true, value: null };
            var existsBySignFromPreviousProcessing = LodashLike.get(object, RecorderConstants.ENTITY_EXISTS_BY_SIGN_FROM_PREVIOUS_PROCESSING);
            if (key === thisLocal.manager.config.playerMetadatasName) {
                customizerResult.value = mdSource;
                customizerResult.needSet = false;
                if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLikeMerge.group('mergeWithCustomizerPropertyReplection => function: (key === thisLocal.manager.config.playerMetadatasName). srcValue and mdSource:');
                    thisLocal.consoleLikeMerge.debug(srcValue);
                    thisLocal.consoleLikeMerge.debug(mdSource);
                    thisLocal.consoleLikeMerge.groupEnd();
                }
                var correctSrcValueAsMetadata = { $iAmPlayerMetadatas$: true };
                Object.assign(correctSrcValueAsMetadata, mdSource);
                if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLikeMerge.debug('mergeWithCustomizerPropertyReplection => function: (key === thisLocal.manager.config.playerMetadatasName). ' +
                        'Immediately assign metadatas on ' +
                        'final instance because resolveMetadatas() is synchronous: \n' +
                        JSON.stringify(correctSrcValueAsMetadata, null, 2));
                }
                if (!existsBySignFromPreviousProcessing) {
                    LodashLike.set(object, key, correctSrcValueAsMetadata);
                }
                if (mdPlayerObjectId.$isComponent$) {
                    customizerResult.value = DummyUndefinedForMergeAsync;
                    //let processResultEntityPrivPlayerObjectId$ = combineFirstSerial(asyncCombineObsArr).pipe(
                    var playerObjectIdValue = thisLocal.processResultEntityPriv(fieldEtc.objectIdPrpType, mdSource.$playerObjectId$, refMap);
                    correctSrcValueAsMetadata.$playerObjectId$ = playerObjectIdValue;
                    customizerResult.value = correctSrcValueAsMetadata;
                    //isDoneRef.result = mdPlayerObjectId;
                    customizerResult.needSet = false;
                    //mdPlayerObjectId.$playerObjectId$ = playerObjectIdValue;
                    if (!existsBySignFromPreviousProcessing) {
                        LodashLike.set(object, key, customizerResult.value);
                    }
                }
            }
            else if (!mdSrcValue.$idRef$ && !isLazyRefField && fieldEtc.fieldProcessorCaller.callFromLiteralValue) {
                customizerResult.value = DummyUndefinedForMergeAsync;
                var callFromLiteralValue = fieldEtc.fieldProcessorCaller.callFromLiteralValue(srcValue, fieldEtc.fieldInfo);
                customizerResult.value = callFromLiteralValue;
                customizerResult.needSet = false;
                if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLikeMerge.debug('(Async) mergeWithCustomizerPropertyReplection => function =>' +
                        ' createSerialAsyncTasksWaiting().pipe() => this.mapJustOnceKeepAllFlagsRxOpr().' +
                        ' Object resolved by fieldEtc.fieldProcessorCaller.callFromLiteralValue:\n' +
                        _this.jsonStringfyWithMax(srcValue));
                }
                if (!existsBySignFromPreviousProcessing) {
                    LodashLike.set(object, key, customizerResult.value);
                }
            }
            else if (mdSrcValue.$idRef$ && !isLazyRefField) {
                customizerResult.value = DummyUndefinedForMergeAsync;
                customizerResult.value = refMap.get(mdSrcValue.$idRef$);
                customizerResult.needSet = false;
                if (!customizerResult.value) {
                    throw new Error('This should not happen 2');
                }
                if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLikeMerge.group('(Async) mergeWithCustomizerPropertyReplection => function =>' +
                        ' createSerialAsyncTasksWaiting().pipe() => this.mapJustOnceKeepAllFlagsRxOpr().' +
                        ' Object resolved by mdSrcValue.$idRef$ field. owner type: ' +
                        object.constructor.name + '; owner field: ' + key);
                    thisLocal.consoleLikeMerge.groupEnd();
                }
                if (!existsBySignFromPreviousProcessing) {
                    LodashLike.set(object, key, customizerResult.value);
                }
            }
            else if (fieldEtc.prpType) {
                var isFromLiteralValue = { value: false };
                if (fieldEtc.prpGenType) {
                    if (fieldEtc.otmCollectionType &&
                        !(fieldEtc.lazyRefMarkerType === LazyRefMTOMarker
                            || fieldEtc.lazyRefMarkerType === LazyRefOTMMarker
                            || fieldEtc.lazyRefMarkerType === LazyRefPrpMarker)) {
                        if (1 === 1) {
                            throw new Error('Not pesrsistent collection is not supported yet!');
                        }
                        if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                            thisLocal.consoleLikeMerge.debug('mergeWithCustomizerPropertyReplection => function.' +
                                ' fieldEtc.otmCollectionType ' + fieldEtc.otmCollectionType.name);
                        }
                        customizerResult.value = DummyUndefinedForMergeAsync;
                        var correctSrcValueColl_1 = thisLocal.createCollection(fieldEtc.otmCollectionType, object, key);
                        thisLocal.lazyLoadTemplateCallback(correctSrcValueColl_1, function () {
                            var entityArr = [];
                            for (var index = 0; index < srcValue.length; index++) {
                                var arrItemType = fieldEtc.lazyLoadedObjType;
                                entityArr.push(thisLocal.processResultEntityPriv(arrItemType, srcValue[index], refMap));
                            }
                            for (var _i = 0, entityArr_1 = entityArr; _i < entityArr_1.length; _i++) {
                                var entityItem = entityArr_1[_i];
                                thisLocal.addOnCollection(correctSrcValueColl_1, entityItem);
                            }
                            customizerResult.value = correctSrcValueColl_1;
                        });
                        //nothing for now
                    }
                    else if (fieldEtc.lazyRefMarkerType === LazyRefOTMMarker
                        || fieldEtc.lazyRefMarkerType === LazyRefMTOMarker
                        || fieldEtc.lazyRefMarkerType === LazyRefPrpMarker) {
                        if (!mdSource.$id$) {
                            throw new Error('There is no mdSource.$id$ on ' + _this.jsonStringfyWithMax(srcValue));
                        }
                        if (mdSrcValueFound && !mdSrcValue.$idRef$ && !mdSrcValue.$isAssociative$ && !mdSrcValue.$isLazyProperty$) {
                            throw new Error('Receiving object that is non associative, no lazy property and has no $idRef$, but field is a LazyRef type. field: ' + object.constructor.name + '.' + key + '. Value' + _this.jsonStringfyWithMax(srcValue));
                        }
                        if (mdSrcValue.$isLazyUninitialized$) {
                            customizerResult.value = DummyUndefinedForMergeAsync;
                            var lazyRef = thisLocal.createNotLoadedLazyRef(fieldEtc.prpGenType, srcValue, refMap, object, key);
                            //for debug purpose
                            srcValue === srcValue;
                            if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                                thisLocal.consoleLikeMerge.group('(Asynchronous of Asynchronous of...) mergeWithCustomizerPropertyReplection =>' +
                                    ' function => createNotLoadedLazyRef$.pipe(thisLocal.mapJustOnceKeepAllFlagsRxOpr()).' +
                                    ' createNotLoadedLazyRef, for property \'' + key + '\'. lodashSet(object, key, lazyRef)');
                                thisLocal.consoleLikeMerge.debug(object);
                                thisLocal.consoleLikeMerge.groupEnd();
                            }
                            customizerResult.value = lazyRef;
                            customizerResult.needSet = true;
                            if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                                thisLocal.consoleLikeMerge.group('mergeWithCustomizerPropertyReplection => function.' +
                                    ' Returning null because of createNotLoadedLazyRef$.subscribe().' +
                                    ' property \'' + key + '\'.');
                                thisLocal.consoleLikeMerge.debug(object);
                                thisLocal.consoleLikeMerge.groupEnd();
                            }
                        }
                        else {
                            customizerResult.value = DummyUndefinedForMergeAsync;
                            var lazyRef = thisLocal.createLoadedLazyRef(fieldEtc, srcValue, refMap, object, key);
                            if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                                thisLocal.consoleLikeMerge.group('(Asynchronous of Asynchronous of...)' +
                                    ' mergeWithCustomizerPropertyReplection => function =>' +
                                    ' createLoadedLazyRef$.subscribe(). createLoadedLazyRef,' +
                                    ' for property \'' + key + '\'. lodashSet(object, key, lazyRef)');
                                thisLocal.consoleLikeMerge.debug(object);
                                thisLocal.consoleLikeMerge.groupEnd();
                            }
                            customizerResult.value = lazyRef;
                            customizerResult.needSet = true;
                            keepAllFlagsTemplateCallback(function () {
                                LodashLike.set(object, key, customizerResult.value);
                            });
                        }
                    }
                    else if (thisLocal.isCollection(fieldEtc.prpType)) {
                        if (!fieldEtc.prpGenType.gParams[0]) {
                            throw new Error('Ivalid non persistent collection with por property ' +
                                object.constructor.name + '.' + key + ':\n' + fieldEtc.prpGenType);
                        }
                        var nonPersistentCollection_1 = _this.createCollection(fieldEtc.prpType, object, key);
                        thisLocal.lazyLoadTemplateCallback(nonPersistentCollection_1, function () {
                            for (var _i = 0, srcValue_1 = srcValue; _i < srcValue_1.length; _i++) {
                                var item = srcValue_1[_i];
                                if (!LodashLike.isNil(item)) {
                                    thisLocal.addOnCollection(nonPersistentCollection_1, thisLocal.processResultEntityPriv(fieldEtc.prpGenType.gParams[0], item, refMap));
                                }
                                else {
                                    thisLocal.addOnCollection(nonPersistentCollection_1, null);
                                }
                            }
                        });
                        customizerResult.value = nonPersistentCollection_1;
                        customizerResult.needSet = true;
                    }
                }
                else if (LodashLike.isObject(srcValue, new Set([Date, Buffer]))
                    && !fieldEtc.propertyOptions.lazyDirectRawRead) {
                    customizerResult.value = DummyUndefinedForMergeAsync;
                    var correctSrcValueSubs_1 = thisLocal.processResultEntityPriv(fieldEtc.prpType, srcValue, refMap);
                    if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                        thisLocal.consoleLikeMerge.group('(Asynchronous of Asynchronous of...) mergeWithCustomizerPropertyReplection =>' +
                            ' function => processResultEntityPriv$.pipe() => thisLocal.mapJustOnceKeepAllFlagsRxOpr().' +
                            ' createLoadedLazyRef, for property \'' + key + '\'. LodashLike.set(object, key, correctSrcValue)');
                        thisLocal.consoleLikeMerge.debug(object);
                        thisLocal.consoleLikeMerge.groupEnd();
                    }
                    customizerResult.value = correctSrcValueSubs_1;
                    customizerResult.needSet = true;
                    keepAllFlagsTemplateCallback(function () {
                        LodashLike.set(object, key, correctSrcValueSubs_1);
                    });
                }
                else if (fieldEtc.fieldProcessorCaller.callFromLiteralValue) {
                    if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                        thisLocal.consoleLikeMerge.group('mergeWithCustomizerPropertyReplection => function.' +
                            ' Transformation by "IFieldProcessor.fromLiteralValue" for property \'' + key + '\'.');
                        thisLocal.consoleLikeMerge.debug(object);
                        thisLocal.consoleLikeMerge.groupEnd();
                    }
                    customizerResult.value = DummyUndefinedForMergeAsync;
                    isFromLiteralValue.value = true;
                    var fromLiteralValue = fieldEtc.fieldProcessorCaller.callFromLiteralValue(srcValue, fieldEtc.fieldInfo);
                    if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                        thisLocal.consoleLikeMerge.group('(Asynchronous of Asynchronous of...) mergeWithCustomizerPropertyReplection =>' +
                            ' function => fromLiteralValue$.pipe() => thisLocal.mapJustOnceKeepAllFlagsRxOpr().' +
                            ' fromLiteralValue, for property \'' + key + '\'. LodashLike.set(object, key, correctSrcValue)');
                        thisLocal.consoleLikeMerge.debug(object);
                        thisLocal.consoleLikeMerge.groupEnd();
                    }
                    customizerResult.value = fromLiteralValue;
                    customizerResult.needSet = false;
                    if (!existsBySignFromPreviousProcessing) {
                        keepAllFlagsTemplateCallback(function () {
                            LodashLike.set(object, key, customizerResult.value);
                        });
                    }
                    if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                        thisLocal.consoleLikeMerge.group('mergeWithCustomizerPropertyReplection => function. Returning null because of fromLiteralValue$.pipe(tap()). property \'' + key + '\'.');
                        thisLocal.consoleLikeMerge.debug(fromLiteralValue);
                        thisLocal.consoleLikeMerge.groupEnd();
                    }
                }
                else {
                    if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                        thisLocal.consoleLikeMerge.group('mergeWithCustomizerPropertyReplection => function. Transformation is not necessary for property \'' + key + '\'.');
                        thisLocal.consoleLikeMerge.debug(object);
                        thisLocal.consoleLikeMerge.groupEnd();
                    }
                    customizerResult.value = srcValue;
                    customizerResult.needSet = false;
                    var correctSrcValue = customizerResult.value;
                    if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                        thisLocal.consoleLikeMerge.group('mergeWithCustomizerPropertyReplection => function. noTranslation$.pipe(thisLocal.mapJustOnceKeepAllFlagsRxOpr()). Transformation is not necessary for property \'' + key + '\'.');
                        thisLocal.consoleLikeMerge.debug(object);
                        thisLocal.consoleLikeMerge.groupEnd();
                    }
                    if (!existsBySignFromPreviousProcessing) {
                        LodashLike.set(object, key, correctSrcValue);
                    }
                }
            }
            else if (LodashLike.has(object, key)) {
                throw new Error('No type decorator for ' + object.constructor.name + '.' + key);
            }
            else if (!LodashLike.has(object, key) && !thisLocal.isLiteralObjMetadataKey(key)) {
                if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLikeMerge.warn('mergeWithCustomizerPropertyReplection => function. This property \'' + key + '\' does not exists on this type.');
                }
                customizerResult.value = undefined;
                customizerResult.needSet = false;
            }
            else {
                if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLikeMerge.group('mergeWithCustomizerPropertyReplection => function. Property \'' + key + '\'. Using same value.');
                    thisLocal.consoleLikeMerge.debug(customizerResult.value);
                    thisLocal.consoleLikeMerge.groupEnd();
                }
            }
            if (thisLocal.consoleLikeMerge.enabledFor(RecorderLogLevel.Trace)) {
                thisLocal.consoleLikeMerge.group('mergeWithCustomizerPropertyReplection => function. return');
                thisLocal.consoleLikeMerge.debug(customizerResult.value);
                thisLocal.consoleLikeMerge.groupEnd();
            }
            if (customizerResult.needSet && existsBySignFromPreviousProcessing) {
                customizerResult.needSet = false;
            }
            return customizerResult;
        };
    };
    RecorderSessionDefault.prototype.registerEntityAndLazyref = function (entity, lazyRef) {
        if (!lazyRef.isLazyLoaded()) {
            throw new Error('Can not register: !lazyRef.isLazyLoaded(). \n' + this.toString());
        }
        if (lazyRef.lazyLoadedObj !== entity) {
            throw new Error('Can not register: entity !== lazyRef.lazyLoadedObj. \n' + this.toString());
        }
        if (!this._lazyrefsByEntityMap.has(entity)) {
            this._lazyrefsByEntityMap.set(entity, new Set());
        }
        this._lazyrefsByEntityMap.get(entity).add(lazyRef);
    };
    RecorderSessionDefault.prototype.unregisterEntityAndLazyref = function (entity, lazyRef) {
        if (!lazyRef.isLazyLoaded()) {
            throw new Error('Can not unregister: !lazyRef.isLazyLoaded(). \n' + this.toString());
        }
        if (lazyRef.lazyLoadedObj === entity) {
            throw new Error('Can not unregister: entity === lazyRef.lazyLoadedObj. \n' + this.toString());
        }
        if (!this._lazyrefsByEntityMap.has(entity)
            || this._lazyrefsByEntityMap.get(entity).size <= 0) {
            throw new Error('Can not unregister: entity has no lazyRef associated. \n' + this.toString());
        }
        this._lazyrefsByEntityMap.get(entity).delete(lazyRef);
    };
    RecorderSessionDefault.prototype.notifyAllLazyrefsAboutEntityModification = function (entity, lazyRefSource) {
        var thisLocal = this;
        if (this._switchedOffNotificationEntitiesSet.has(entity)) {
            if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                thisLocal.consoleLike.group('RecorderSessionDefault.notifyLazyrefAboutEntityModification: this._switchOffEntitiesSet.has(entity): entity modified but modifying notification is switched off!. entity:');
                thisLocal.consoleLike.debug(entity);
                thisLocal.consoleLike.groupEnd();
            }
        }
        else {
            if (!lazyRefSource) {
                if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLike.group('RecorderSessionDefault.notifyLazyrefAboutEntityModification: !lazyRefSource: Am I been notified from a property set?!. entity:');
                    thisLocal.consoleLike.debug(entity);
                    thisLocal.consoleLike.groupEnd();
                }
            }
            if (lazyRefSource && !lazyRefSource.isLazyLoaded()) {
                throw new Error('Can not notify: !lazyRefSource.isLazyLoaded(). \n' + this.toString());
            }
            if (lazyRefSource && lazyRefSource.lazyLoadedObj !== entity) {
                throw new Error('Can not notify: entity !== lazyRefSource.lazyLoadedObj. \n' + this.toString());
            }
            if (this._lazyrefsByEntityMap.has(entity)) {
                var lazyrefsArr = Array.from(this._lazyrefsByEntityMap.get(entity));
                for (var _i = 0, lazyrefsArr_1 = lazyrefsArr; _i < lazyrefsArr_1.length; _i++) {
                    var lazyrefItem = lazyrefsArr_1[_i];
                    if (lazyRefSource === lazyrefItem) {
                        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                            thisLocal.consoleLike.group('RecorderSessionDefault.notifyLazyrefAboutEntityModification: Not notifing: lazyRefSource === lazyrefItem. entity:');
                            thisLocal.consoleLike.debug(entity);
                            thisLocal.consoleLike.groupEnd();
                        }
                    }
                    else {
                        if (!lazyrefItem.isLazyLoaded()) {
                            throw new Error('Can not notify: !lazyrefItem.isLazyLoaded(). \n' + this.toString());
                        }
                        if (lazyrefItem.lazyLoadedObj !== entity) {
                            throw new Error('Can not notify: lazyrefItem.lazyLoadedObj !== entity. \n' + this.toString());
                        }
                        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                            thisLocal.consoleLike.group('RecorderSessionDefault.notifyLazyrefAboutEntityModification: Notifing: lazyRefSource !== lazyrefItem.\n lazyRefSource and lazyrefItem:\n' + lazyRefSource + '\n' + lazyrefItem + '\n an . entity:');
                            thisLocal.consoleLike.debug(entity);
                            thisLocal.consoleLike.groupEnd();
                        }
                        lazyrefItem.notifyModification(entity);
                    }
                }
            }
        }
    };
    RecorderSessionDefault.prototype.switchOffNotifyAllLazyrefs = function (entity) {
        this._switchedOffNotificationEntitiesSet.add(entity);
    };
    RecorderSessionDefault.prototype.switchOnNotifyAllLazyrefs = function (entity) {
        this._switchedOffNotificationEntitiesSet.delete(entity);
    };
    /** Framework internal use. */
    RecorderSessionDefault.prototype.nextMultiPurposeInstanceId = function () {
        return this._nextMultiPurposeInstanceId++;
    };
    RecorderSessionDefault.prototype.processTapeActionAttachRefId = function (options) {
        var _this = this;
        var thisLocal = this;
        var resultObservableValue = {
            asyncAddTapeAction: false,
            newValue: undefined
        };
        var resultObservableValue$ = of(resultObservableValue);
        if (!options.action.attachRefId) {
            //non LazyRef lazyDirectRawWrite fields has not unchengable attachRefId.
            options.action.attachRefId = thisLocal.manager.config.cacheStoragePrefix + thisLocal.nextMultiPurposeInstanceId();
        }
        if (options.fieldEtc.fieldProcessorCaller && options.fieldEtc.fieldProcessorCaller.callToDirectRaw) {
            var toDirectRaw$ = options.fieldEtc.fieldProcessorCaller.callToDirectRaw(options.value, options.fieldEtc.fieldInfo);
            // toDirectRaw$ = toDirectRaw$.pipe(thisLocal.addSubscribedObsRxOpr());
            resultObservableValue.asyncAddTapeAction = true;
            resultObservableValue$ = toDirectRaw$.pipe(flatMap(function (respStream) {
                var putOnCacheGetFromCache$ = of(undefined);
                if (respStream.body) {
                    putOnCacheGetFromCache$ = thisLocal.manager.config.cacheHandler.putOnCache(options.action.attachRefId, respStream.body);
                    // putOnCacheGetFromCache$ = putOnCacheGetFromCache$.pipe(thisLocal.addSubscribedObsRxOpr());
                    putOnCacheGetFromCache$ = putOnCacheGetFromCache$.pipe(flatMap(function () {
                        thisLocal.addTapeAction(options.action);
                        var getFromCache$ = thisLocal.manager.config.cacheHandler.getFromCache(options.action.attachRefId);
                        // getFromCache$ = getFromCache$.pipe(thisLocal.addSubscribedObsRxOpr());
                        getFromCache$ = getFromCache$.pipe(tap(function (stream) {
                            // oldSet.call(this, stream);
                            resultObservableValue.newValue = stream;
                        }));
                        return getFromCache$;
                    }), map(function (stream) {
                        resultObservableValue.newValue = stream;
                        return resultObservableValue;
                    }));
                    return putOnCacheGetFromCache$;
                }
                else {
                    if (options.value) {
                        throw new Error('The property \'' + options.propertyKey.toString() + ' of \'' + _this.constructor + '\'. NodeJS.ReadableStream is null but value is not null. value: ' + options.value.constructor);
                    }
                    options.action.simpleSettedValue = null;
                    options.action.attachRefId = null;
                    thisLocal.addTapeAction(options.action);
                    return of(resultObservableValue);
                }
            }), share());
            return resultObservableValue$;
        }
        else {
            if (!(options.value.addListener && options.value.pipe)) {
                throw new Error('The property \'' + options.propertyKey.toString() +
                    ' of \'' + this.constructor + '\'. There is no "IFieldProcessor.toDirectRaw"' +
                    ' defined and value is not a NodeJS.ReadableStream. value: ' + options.value.constructor);
            }
            else {
                var putOnCacheGetFromCache$ = thisLocal.manager.config.cacheHandler.putOnCache(options.action.attachRefId, options.value);
                // putOnCacheGetFromCache$ = putOnCacheGetFromCache$.pipe(thisLocal.addSubscribedObsRxOpr());
                resultObservableValue.asyncAddTapeAction = true;
                resultObservableValue$ = putOnCacheGetFromCache$.pipe(tap(function () {
                    thisLocal.addTapeAction(options.action);
                }), flatMap(function () {
                    var getFromCache$ = thisLocal.manager.config.cacheHandler.getFromCache(options.action.attachRefId);
                    // getFromCache$ = getFromCache$.pipe(thisLocal.addSubscribedObsRxOpr());
                    getFromCache$ = getFromCache$.pipe(tap({
                        next: function (stream) {
                            //oldSet.call(this, stream);
                            resultObservableValue.newValue = stream;
                        }
                    }));
                    return getFromCache$;
                }), map(function () {
                    return resultObservableValue;
                }), share());
            }
        }
    };
    RecorderSessionDefault.prototype.registerLazyRefSubscriptionRxOpr = function (signature) {
        var thisLocal = this;
        //BEGIN: Used to find losted obs.subscribed()
        var stackSubscriberRef = { value: '' };
        try {
            throw new Error('TRACKING');
        }
        catch (error) {
            stackSubscriberRef.value = error.stack;
        }
        ;
        //END: Used to find losted obs.subscribed()
        var resultOpr = function (source) {
            var sourceRef = { value: source };
            //thisLocal.testObservableHasCircle(source);
            var result$ = source;
            var pendingLazyRefSubscription = thisLocal.pendingLazyRefSubscriptionBySign.get(signature);
            if (pendingLazyRefSubscription) {
                result$ = pendingLazyRefSubscription;
            }
            else {
                result$ = of(null).pipe(tap({
                    next: function () {
                        thisLocal.pendingLazyRefSubscriptionBySign.set(signature, source);
                    },
                    error: function () {
                        thisLocal.pendingLazyRefSubscriptionBySign.set(signature, source);
                    }
                }), flatMap(function () {
                    sourceRef.value = source;
                    return source;
                }), tap({
                    next: function () {
                        thisLocal.pendingLazyRefSubscriptionBySign.delete(signature);
                    },
                    error: function () {
                        thisLocal.pendingLazyRefSubscriptionBySign.delete(signature);
                    }
                })
                // ,
                // timeoutDecorateRxOpr()
                );
            }
            return result$;
        };
        return resultOpr;
    };
    RecorderSessionDefault.prototype.jsonStringfyWithMax = function (literalObj) {
        var result = '';
        if (literalObj) {
            var literalObjStr = JSON.stringify(literalObj, null, 2);
            if (literalObjStr.length > this.manager.config.maxJsonStringifyForDiagnostic) {
                result += literalObjStr.substr(0, this.manager.config.maxJsonStringifyForDiagnostic) + '\n...';
            }
            else {
                result += literalObjStr.substr(0, this.manager.config.maxJsonStringifyForDiagnostic) + '\n...';
            }
        }
        return result;
    };
    return RecorderSessionDefault;
}());
export { RecorderSessionDefault };
//# sourceMappingURL=recorder-session-default.js.map