import { TapeActionType } from '../api/tape';
import { RecorderConstants } from './recorder-constants';
import { RecorderLogger, RecorderLogLevel } from '../api/recorder-config';
import { TapeActionDefault } from './tape-default';
import { LodashLike } from './lodash-like';
var SetCreator = /** @class */ (function () {
    function SetCreator(_session, _refererObj, _refererKey) {
        this._session = _session;
        this._refererObj = _refererObj;
        this._refererKey = _refererKey;
        var thisLocal = this;
        if (!_session) {
            throw new Error('_session can not be null');
        }
        if (!_refererObj) {
            throw new Error('_refererObj can not be null');
        }
        if (!_session) {
            throw new Error('_session can not be null');
        }
        thisLocal.consoleLike = _session.manager.config.getConsole(RecorderLogger.SetCreator);
        if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
            thisLocal.consoleLike.group('SetCreator.constructor()');
            thisLocal.consoleLike.debug(_session);
            thisLocal.consoleLike.debug(_refererObj);
            thisLocal.consoleLike.debug(_refererKey);
            thisLocal.consoleLike.groupEnd();
        }
    }
    SetCreator.prototype.createByProxy = function () {
        var _this = this;
        var thisLocal = this;
        var getFunction = function (target, p) {
            if (p) {
                if (thisLocal.consoleLike.enabledFor(RecorderLogLevel.Trace)) {
                    thisLocal.consoleLike.group('SetCreator => getFunction. Intercepting:');
                    thisLocal.consoleLike.debug(target);
                    thisLocal.consoleLike.debug(p);
                    //thisLocal.consoleLike.debug(receiver);
                    thisLocal.consoleLike.groupEnd();
                }
                if (p === 'add') {
                    var proxyGet = function (value) {
                        _this.add(target, value);
                        return target.add(value);
                    };
                    return proxyGet;
                }
                else if (p === 'delete') {
                    var proxyDelete = function (value) {
                        _this.add(target, value);
                        return target.delete(value);
                    };
                    return proxyDelete;
                }
                else if (p === 'clear') {
                    var proxyClear = function () {
                        target.forEach(function (value) {
                            _this.delete(target, value);
                        });
                        target.clear();
                    };
                    return proxyClear;
                }
                else if (target[p] && target[p] instanceof Function) {
                    return function (argArray) {
                        return target[p](argArray);
                    };
                }
                else {
                    return target[p];
                }
            }
            else {
                return undefined;
            }
        };
        return new Proxy(new Set(), {
            get: getFunction
        });
    };
    SetCreator.prototype.add = function (targetSet, value) {
        var propertyOptions = Reflect.getMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_OBJECT_PROPERTY_OPTIONS, this.refererObj, this.refererKey);
        if (!propertyOptions) {
            throw new Error('@RecorderDecorators.property() not defined for ' + this.refererObj.constructor.name + '.' + this.refererKey);
        }
        if (propertyOptions.persistent) {
            var isOnlazyLoad = LodashLike.get(targetSet, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME);
            if (!this.session.isOnRestoreEntireStateFromLiteral() && !isOnlazyLoad) {
                if (!this.session.isRecording()) {
                    throw new Error('Invalid operation. It is not recording. Is this Error correct?!');
                }
                var allMD = this.session.resolveMetadatas({ object: value, refererObject: this.refererObj, key: this.refererKey });
                var mdRefererObj = allMD.refererObjMd;
                var mdValue = allMD.objectMd;
                //recording tape
                var action = new TapeActionDefault();
                action.fieldName = this.refererKey;
                action.actionType = TapeActionType.CollectionAdd;
                if (mdRefererObj.$signature$) {
                    action.ownerSignatureStr = mdRefererObj.$signature$;
                }
                else if (LodashLike.has(this.refererObj, this.session.manager.config.creationIdName)) {
                    action.ownerCreationRefId = LodashLike.get(this.refererObj, this.session.manager.config.creationIdName);
                }
                else if (!mdRefererObj.$isComponentPlayerObjectId$) {
                    throw new Error('The proprerty \'' + this.refererKey + ' from \'' + this.refererObj.constructor.name + '\' has a not managed owner');
                }
                if (value != null) {
                    if (mdValue.$signature$) {
                        action.settedSignatureStr = mdValue.$signature$;
                    }
                    else if (LodashLike.has(value, this.session.manager.config.creationIdName)) {
                        action.settedCreationRefId = LodashLike.get(value, this.session.manager.config.creationIdName);
                    }
                    else {
                        throw new Error('The proprerty \'' + this.refererKey + ' of \'' + this.refererObj.constructor.name + '\'.  value not managed owner: \'' + value.constructor.name + '\'');
                    }
                }
                this.session.addTapeAction(action);
            }
        }
    };
    SetCreator.prototype.delete = function (targetSet, value) {
        var propertyOptions = Reflect.getMetadata(RecorderConstants.REFLECT_METADATA_PLAYER_OBJECT_PROPERTY_OPTIONS, this.refererObj, this.refererKey);
        if (!propertyOptions) {
            throw new Error('@RecorderDecorators.property() not defined for ' + this.refererObj.constructor.name + '.' + this.refererKey);
        }
        if (propertyOptions.persistent) {
            var isOnlazyLoad = LodashLike.get(targetSet, RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME);
            if (!this.session.isOnRestoreEntireStateFromLiteral() && !isOnlazyLoad) {
                if (!this.session.isRecording()) {
                    throw new Error('Invalid operation. It is not recording. Is this Error correct?!');
                }
                var allMD = this.session.resolveMetadatas({ object: value, refererObject: this.refererObj, key: this.refererKey });
                var mdRefererObj = allMD.refererObjMd;
                var mdValue = allMD.objectMd;
                //recording tape
                var action = new TapeActionDefault();
                action.fieldName = this.refererKey;
                action.actionType = TapeActionType.CollectionRemove;
                if (mdRefererObj.$signature$) {
                    action.ownerSignatureStr = mdRefererObj.$signature$;
                }
                else if (LodashLike.has(this.refererObj, this.session.manager.config.creationIdName)) {
                    action.ownerCreationRefId = LodashLike.get(this.refererObj, this.session.manager.config.creationIdName);
                }
                else {
                    throw new Error('The proprerty \'' + this.refererKey + ' of \'' + this.refererObj.constructor + '\' has a not managed owner');
                }
                if (value != null) {
                    if (mdValue.$signature$) {
                        action.settedSignatureStr = mdValue.$signature$;
                    }
                    else if (LodashLike.has(value, this.session.manager.config.creationIdName)) {
                        action.settedCreationRefId = LodashLike.get(value, this.session.manager.config.creationIdName);
                    }
                    else {
                        throw new Error('The proprerty \'' + this.refererKey + ' of \'' + this.refererObj.constructor + '\'. not managed value: \'' + value.constructor.name + '\'');
                    }
                }
                this.session.addTapeAction(action);
            }
        }
    };
    Object.defineProperty(SetCreator.prototype, "session", {
        /**
         * Getter session
         * @return {RecorderSession}
         */
        get: function () {
            return this._session;
        },
        /**
         * Setter session
         * @param {RecorderSession} value
         */
        set: function (value) {
            this._session = value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SetCreator.prototype, "refererObj", {
        /**
         * Getter refererObj
         * @return {any}
         */
        get: function () {
            return this._refererObj;
        },
        /**
         * Setter refererObj
         * @param {any} value
         */
        set: function (value) {
            this._refererObj = value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SetCreator.prototype, "refererKey", {
        /**
         * Getter refererKey
         * @return {string}
         */
        get: function () {
            return this._refererKey;
        },
        /**
         * Setter refererKey
         * @param {string} value
         */
        set: function (value) {
            this._refererKey = value;
        },
        enumerable: true,
        configurable: true
    });
    return SetCreator;
}());
export { SetCreator };
//# sourceMappingURL=set-creator.js.map