/// <reference types="node" />
import { Observable, OperatorFunction, ObservableInput, Subject, PartialObserver, Subscription } from 'rxjs';
import { ResponseLike } from '../typeslike';
import { LazyRefImplementor } from '../api/lazy-ref';
import { GenericNode } from '../api/generic-tokenizer';
import { IFieldProcessorEvents } from '../api/field-processor';
import { RecorderSessionImplementor, PlayerSnapshot } from '../api/recorder-session';
import { PlayerMetadatas } from '../api/player-metadatas';
import { FieldEtc } from '../api/field-etc';
import { ConsoleLike } from '../api/recorder-config';
export declare abstract class LazyRefBase<L extends object, I> extends Subject<L> implements LazyRefImplementor<L, I> {
    iAmLazyRef: true;
    iAmLazyRefImplementor: true;
    attachRefId: string;
    bMdRefererObj: PlayerMetadatas;
    bMdLazyLoadedObj: PlayerMetadatas;
    pbMdRefererPlayerObjectId: PlayerMetadatas;
    private notificationStartTime;
    private notificationCount;
    private _instanceId;
    instanceId: number;
    private _lazyLoadedObj;
    private _genericNode;
    private _signatureStr;
    private _respObs;
    private _refererObj;
    private _refererKey;
    private _session;
    consoleLike: ConsoleLike;
    consoleLikeProcResp: ConsoleLike;
    consoleLikeSubs: ConsoleLike;
    constructor(session: RecorderSessionImplementor);
    private _isOnLazyLoading;
    protected readonly isOnLazyLoading: boolean;
    protected flatMapKeepAllFlagsRxOprPriv<T, R>(when: 'justOnce' | 'eachPipe', project: (value: T, index?: number) => ObservableInput<R>, concurrent?: number): OperatorFunction<T, R>;
    protected flatMapKeepAllFlagsRxOpr<T, R>(project: (value: T, index?: number) => ObservableInput<R>): OperatorFunction<T, R>;
    protected flatMapJustOnceKeepAllFlagsRxOpr<T, R>(project: (value: T, index?: number) => ObservableInput<R>): OperatorFunction<T, R>;
    protected mapKeepAllFlagsRxOprHelper<T, R>(when: 'justOnce' | 'eachPipe', project: (value: T, index?: number) => R): OperatorFunction<T, R>;
    protected mapJustOnceKeepAllFlagsRxOpr<T, R>(project: (value: T, index?: number) => R): OperatorFunction<T, R>;
    setLazyObjOnLazyLoading(lazyLoadedObj: L, observerOriginal: PartialObserver<L>): void;
    setLazyObjOnLazyLoadingProt(lazyLoadedObj: L, observerOriginal: PartialObserver<L>): void;
    isLazyLoaded(): boolean;
    private _needCallNextOnSetLazyObj;
    setLazyObjOnLazyLoadingNoNext(lazyLoadedObj: L): void;
    setLazyObjNoNext(lazyLoadedObj: L): void;
    notifyModification(lazyLoadedObj: L): void;
    protected processResponseOnLazyLoading(responselike: {
        body: any;
    }): L | Observable<L>;
    private lazyLoadingCallbackTemplate;
    private noNextCallbackTemplate;
    protected setLazyObjInternal(lazyLoadedObj: L, observerOriginal: PartialObserver<L>): void;
    abstract setLazyObj(lazyLoadedObj: L, observerOriginal?: PartialObserver<L>): void;
    protected setLazyObjMayDoNextHelper(isValueByFieldProcessor: {
        value: boolean;
    }, fieldEtc: FieldEtc<L, any>, newLazyLoadedObj: L, observerOriginal: PartialObserver<L>): void;
    private _subscribeFromAsObservable;
    protected subscribeFromAsObservable: boolean;
    asObservable(): Observable<L>;
    private DummyCurrNextValueClass;
    protected readonly dummyCurrtNextValueInstance: L;
    protected currNextValue: L;
    protected firstSubscribeOcurrence: boolean;
    private _observerOriginalMap;
    private _observerOriginalInverseMap;
    protected createAndRegisterObserverOriginal(observerOrNext?: PartialObserver<L> | ((value: L) => void), error?: (error: any) => void, complete?: () => void): PartialObserver<L>;
    /**
     * Do super.next only if is the first subscribe ocurred or is the value
     * is diferent from the last value or is forceSuper.
     * @param forceSuper
     * @param value
     */
    protected nextProt(forceSuper: boolean, value: L, observerOriginal: PartialObserver<L>): void;
    next(value?: L): void;
    subscribe(observerOrNext?: PartialObserver<L> | ((value: L) => void), error?: (error: any) => void, complete?: () => void): Subscription;
    subscribeToModify(observerOrNext?: PartialObserver<L> | ((value: L) => void), error?: (error: any) => void, complete?: () => void): void;
    private subscriptionToChange;
    protected subscriptionToChangeUnsubscribe(): void;
    abstract processResponse(responselike: ResponseLike<PlayerSnapshot | NodeJS.ReadStream>): L | Observable<L>;
    protected lazyRefPrpStoreOriginalliteralEntryIfNeeded(mdRefererObj: PlayerMetadatas, fieldEtc: FieldEtc<L, any>, playerSnapshot: PlayerSnapshot): void;
    lazyLoadedObj: L;
    signatureStr: string;
    respObs: Observable<ResponseLike<Object>>;
    private _fieldProcessorEvents;
    /** Framework internal use. */
    readonly fieldProcessorEvents: IFieldProcessorEvents<L>;
    refererObj: any;
    refererKey: string;
    session: RecorderSessionImplementor;
    genericNode: GenericNode;
    toString(): string;
}
