(function (factory) {
    if (typeof module === "object" && typeof module.exports === "object") {
        var v = factory(require, exports);
        if (v !== undefined) module.exports = v;
    }
    else if (typeof define === "function" && define.amd) {
        define(["require", "exports", "util"], factory);
    }
})(function (require, exports) {
    Object.defineProperty(exports, "__esModule", { value: true });
    var util_1 = require("util");
    var LodashLike;
    (function (LodashLike) {
        function isObject(value, noObjects) {
            if (!noObjects) {
                noObjects = new Set();
            }
            var type = typeof value;
            return !isNil(value) != null &&
                (type == 'object' || type == 'function') &&
                (value.constructor && !noObjects.has(value.constructor));
        }
        LodashLike.isObject = isObject;
        function isNil(value) {
            return value === null || value === undefined;
        }
        LodashLike.isNil = isNil;
        function testPatterns(valueToTest, pttrs) {
            for (var index = 0; index < pttrs.length; index++) {
                var pttrItem = pttrs[index];
                if (pttrItem.test(valueToTest)) {
                    return true;
                }
            }
            return false;
        }
        function asyncMergeWithDeep(object, source, visitedMap, 
        //setterResults: AsyncCustomSetterResult[],
        extraOptions) {
            // console.log(object.constructor.name);
            if (!extraOptions) {
                extraOptions = {};
            }
            if (Array.isArray(source)) {
                throw new Error('Not supported. source is array.');
            }
            if (object === source) {
                return object;
            }
            //const allObsArr: Observable<any>[] = [];
            var prpsSet = new Set();
            var finalPrpsArr = [];
            for (var propt in source) {
                prpsSet.add(propt);
            }
            if (extraOptions.considerObjectProperties && !util_1.isNull(object)) {
                for (var propt in object) {
                    prpsSet.add(propt);
                }
            }
            for (var _i = 0, _a = Array.from(prpsSet); _i < _a.length; _i++) {
                var prpItem = _a[_i];
                if (!extraOptions.ignorePropeties || !testPatterns(prpItem, extraOptions.ignorePropeties)) {
                    finalPrpsArr.push(prpItem);
                }
            }
            for (var _b = 0, finalPrpsArr_1 = finalPrpsArr; _b < finalPrpsArr_1.length; _b++) {
                var propt = finalPrpsArr_1[_b];
                //for debug purpose
                finalPrpsArr === finalPrpsArr;
                var prpSourceValue = source[propt];
                var prpObjectValue = !isNil(object) ? object[propt] : undefined;
                var proptMergeResult = void 0;
                var newValue = void 0;
                if (isObject(prpSourceValue, extraOptions.noObjects)) {
                    if (!visitedMap.has(prpSourceValue)) {
                        if (extraOptions.customizer) {
                            proptMergeResult = extraOptions.customizer(prpObjectValue, prpSourceValue, propt, object, source);
                            if (isNil(proptMergeResult)) {
                                throw new Error('Customizer can not return null!');
                            }
                            newValue = proptMergeResult.value;
                            visitedMap.set(prpSourceValue, proptMergeResult.value);
                        }
                        else {
                            newValue = asyncMergeWithDeep(prpObjectValue, prpSourceValue, visitedMap, extraOptions);
                            visitedMap.set(prpSourceValue, newValue);
                        }
                    }
                    else {
                        newValue = visitedMap.get(prpSourceValue);
                    }
                }
                else {
                    if (extraOptions.customizer) {
                        proptMergeResult = extraOptions.customizer(prpObjectValue, prpSourceValue, propt, object, source);
                        if (isNil(proptMergeResult)) {
                            throw new Error('Customizer can not return null!');
                        }
                        newValue = proptMergeResult.value;
                    }
                    else {
                        newValue = asyncMergeWithDeep(prpObjectValue, prpSourceValue, visitedMap, extraOptions);
                    }
                }
                if (proptMergeResult) {
                    if (proptMergeResult.needSet) {
                        set(object, propt, newValue);
                    }
                }
            }
            return object;
        }
        function mergeWithDeep(object, source, customizer, visitedMap, noObjects) {
            if (Array.isArray(source)) {
                throw new Error('Not supported. source is array.');
            }
            if (object === source) {
                return object;
            }
            //const allObsArr: Observable<any>[] = [];
            for (var propt in source) {
                var prpSourceValue = source[propt];
                var prpObjectValue = object[propt];
                var newValue = prpSourceValue;
                if (isObject(prpSourceValue, noObjects)) {
                    if (!visitedMap.has(prpSourceValue)) {
                        if (customizer) {
                            newValue = customizer(prpObjectValue, prpSourceValue, propt, object, source);
                        }
                        else {
                            newValue = mergeWithDeep(prpObjectValue, prpSourceValue, customizer, visitedMap, noObjects);
                        }
                        visitedMap.set(prpSourceValue, newValue);
                    }
                    else {
                        newValue = visitedMap.get(prpSourceValue);
                    }
                }
                else {
                    if (customizer) {
                        newValue = customizer(prpObjectValue, prpSourceValue, propt, object, source);
                    }
                    else {
                        newValue = mergeWithDeep(prpObjectValue, prpSourceValue, customizer, visitedMap, noObjects);
                    }
                }
                set(object, propt, newValue);
            }
            return object;
            // return thisLocal.combineFirstSerialPreserveAllFlags(allObsArr).pipe(
            //     map(() => {
            //     })
            // );
        }
        function mergeWith(object, source, customizer, noObjects) {
            return mergeWithDeep(object, source, customizer, new Map());
        }
        LodashLike.mergeWith = mergeWith;
        function asyncMergeWith(object, source, extraOptions) {
            return asyncMergeWithDeep(object, source, new Map(), extraOptions);
        }
        LodashLike.asyncMergeWith = asyncMergeWith;
        function has(obj, prop) {
            if (!isNil(obj)) {
                var proto = obj.__proto__ || obj.constructor.prototype;
                return isObject(obj) && (prop in obj) &&
                    (!(prop in proto) || proto[prop] !== obj[prop]);
            }
            else {
                return false;
            }
        }
        LodashLike.has = has;
        function get(obj, prop) {
            return !isNil(obj) ? obj[prop] : undefined;
        }
        LodashLike.get = get;
        function set(obj, prop, value) {
            if (!isNil(obj)) {
                obj[prop] = value;
            }
        }
        LodashLike.set = set;
        function clone(source, noObjects) {
            var result = null;
            if (!isNil(source)) {
                if (isObject)
                    result = Object.assign({}, source);
            }
            return result;
        }
        LodashLike.clone = clone;
        function keys(object) {
            return isArrayLike(object)
                ? arrayLikeKeys(object)
                : Object.keys(Object(object));
        }
        LodashLike.keys = keys;
        function isArrayLike(value) {
            return !isNil(value) && typeof value != 'function' && isLength(value.length);
        }
        LodashLike.isArrayLike = isArrayLike;
        function isLength(value) {
            return typeof value == 'number' &&
                value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER;
        }
        var hasOwnProperty = Object.prototype.hasOwnProperty;
        function arrayLikeKeys(value, inherited) {
            var isArr = Array.isArray(value);
            var isArg = !isArr && isArguments(value);
            //const isBuff = !isArr && !isArg && isBuffer(value);
            //const isType = !isArr && !isArg && !isBuff && isTypedArray(value)
            //const skipIndexes = isArr || isArg || isBuff || isType
            var skipIndexes = isArr || isArg;
            var length = value.length;
            var result = new Array(skipIndexes ? length : 0);
            var index = skipIndexes ? -1 : length;
            while (++index < length) {
                result[index] = "" + index;
            }
            for (var key in value) {
                if ((inherited || hasOwnProperty.call(value, key)) &&
                    !(skipIndexes && (
                    // Safari 9 has enumerable `arguments.length` in strict mode.
                    (key == 'length' ||
                        // Skip index properties.
                        isIndex(key, length))))) {
                    result.push(key);
                }
            }
            return result;
        }
        function isArguments(value) {
            return isObject(value) && getTag(value) == '[object Arguments]';
        }
        function getTag(value) {
            if (value == null) {
                return value === undefined ? '[object Undefined]' : '[object Null]';
            }
            return toString.call(value);
        }
        var MAX_SAFE_INTEGER = 9007199254740991;
        var reIsUint = /^(?:0|[1-9]\d*)$/;
        function isIndex(value, length) {
            var type = typeof value;
            length = length == null ? MAX_SAFE_INTEGER : length;
            return !!length &&
                (type == 'number' ||
                    (type != 'symbol' && reIsUint.test(value))) &&
                (value > -1 && value % 1 == 0 && value < length);
        }
    })(LodashLike = exports.LodashLike || (exports.LodashLike = {}));
});
//# sourceMappingURL=lodash-like.js.map