(function (factory) {
    if (typeof module === "object" && typeof module.exports === "object") {
        var v = factory(require, exports);
        if (v !== undefined) module.exports = v;
    }
    else if (typeof define === "function" && define.amd) {
        define(["require", "exports", "../api/recorder-decorators", "../api/recorder-config"], factory);
    }
})(function (require, exports) {
    Object.defineProperty(exports, "__esModule", { value: true });
    var recorder_decorators_1 = require("../api/recorder-decorators");
    var recorder_config_1 = require("../api/recorder-config");
    var ConsoleLikeBase = /** @class */ (function () {
        function ConsoleLikeBase(logger, level) {
            this.logger = logger;
            this.level = level;
        }
        ConsoleLikeBase.prototype.group = function () {
            var label = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                label[_i] = arguments[_i];
            }
            var labelNew = label.slice();
            if (labelNew.length > 0) {
                labelNew[0] = '[' + this.logger + '] ' + labelNew[0];
            }
            console.group.apply(console, labelNew);
        };
        ConsoleLikeBase.prototype.groupEnd = function () {
            console.groupEnd();
        };
        ConsoleLikeBase.prototype.error = function (message) {
            var optionalParams = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                optionalParams[_i - 1] = arguments[_i];
            }
            console.error.apply(console, ['[' + this.logger + '] ' + (message && message.toString ? message.toString() : '')].concat(optionalParams));
        };
        ConsoleLikeBase.prototype.warn = function (message) {
            var optionalParams = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                optionalParams[_i - 1] = arguments[_i];
            }
            console.warn.apply(console, ['[' + this.logger + '] ' + (message && message.toString ? message.toString() : '')].concat(optionalParams));
        };
        ConsoleLikeBase.prototype.log = function (message) {
            var optionalParams = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                optionalParams[_i - 1] = arguments[_i];
            }
            console.log.apply(console, ['[' + this.logger + '] ' + (message && message.toString ? message.toString() : '')].concat(optionalParams));
        };
        ConsoleLikeBase.prototype.debug = function (message) {
            var optionalParams = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                optionalParams[_i - 1] = arguments[_i];
            }
            console.debug.apply(console, ['[' + this.logger + '] ' + (message && message.toString ? message.toString() : '')].concat(optionalParams));
        };
        ConsoleLikeBase.prototype.info = function (message) {
            var optionalParams = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                optionalParams[_i - 1] = arguments[_i];
            }
            console.info.apply(console, ['[' + this.logger + '] ' + (message && message.toString ? message.toString() : '')].concat(optionalParams));
        };
        ConsoleLikeBase.prototype.enabledFor = function (level) {
            return level >= this.level;
        };
        ConsoleLikeBase.prototype.getLevel = function () {
            return this.level;
        };
        return ConsoleLikeBase;
    }());
    exports.ConsoleLikeBase = ConsoleLikeBase;
    var RecorderConfigDefault = /** @class */ (function () {
        function RecorderConfigDefault() {
            this._logLevelMap = new Map();
            this._attachPrefix = 'attachPrefix_';
            this._cacheStoragePrefix = 'cacheStoragePrefix_';
            this._creationIdName = 'creationId';
            this._playerMetadatasName = '$metadatas$';
            this._maxLazyRefNotificationPerSecond = 10;
            this._lazyRefNotificationTimeMeasurement = 5000;
            this._lazyRefNotificationCountMeasurement = 30;
            this._fieldProcessorEntryMap = new Map();
            this._maxJsonStringifyForDiagnostic = 300;
            this._tryReduceLazyRefSubescribersRerun = true;
            this.configAddFieldProcessors(recorder_decorators_1.RecorderDecorators.TypeProcessorEntries);
            this.configCacheHandler({
                clearCache: function () {
                    throw new Error('CacheHandler not defined!');
                },
                getFromCache: function () {
                    throw new Error('CacheHandler not defined!');
                },
                putOnCache: function () {
                    throw new Error('CacheHandler not defined!');
                },
                removeFromCache: function () {
                    throw new Error('CacheHandler not defined!');
                }
            });
        }
        RecorderConfigDefault.prototype.configLazyObservableProvider = function (provider) {
            this.lazyObservableProvider = provider;
            return this;
        };
        RecorderConfigDefault.prototype.configLogLevel = function (logger, level, consoleLike) {
            if (!consoleLike) {
                consoleLike = new ConsoleLikeBase(logger, level);
            }
            this._logLevelMap.set(logger, consoleLike);
            return this;
        };
        RecorderConfigDefault.prototype.getConsole = function (logger) {
            if (!this._logLevelMap.has(logger)) {
                if (this._logLevelMap.has(recorder_config_1.RecorderLogger.All)) {
                    var consoleAll = this._logLevelMap.get(recorder_config_1.RecorderLogger.All);
                    this._logLevelMap.set(logger, new ConsoleLikeBase(logger, consoleAll.getLevel()));
                }
                else {
                    this._logLevelMap.set(logger, new ConsoleLikeBase(logger, recorder_config_1.RecorderLogLevel.Error));
                }
            }
            return this._logLevelMap.get(logger);
        };
        RecorderConfigDefault.prototype.configAttachPrefix = function (attachPrefix) {
            this._attachPrefix = attachPrefix;
            return this;
        };
        RecorderConfigDefault.prototype.configCacheStoragePrefix = function (cacheStoragePrefix) {
            this._cacheStoragePrefix = cacheStoragePrefix;
            return this;
        };
        Object.defineProperty(RecorderConfigDefault.prototype, "attachPrefix", {
            get: function () {
                return this._attachPrefix;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(RecorderConfigDefault.prototype, "cacheStoragePrefix", {
            get: function () {
                return this._cacheStoragePrefix;
            },
            enumerable: true,
            configurable: true
        });
        /**
         * Max size of JSON.stringfy for error and log purpose. Default: 300.
         * @param maxJsonStringifyForDiagnostic
         */
        RecorderConfigDefault.prototype.configMaxJsonStringifyForDiagnostic = function (maxJsonStringifyForDiagnostic) {
            this._maxJsonStringifyForDiagnostic = maxJsonStringifyForDiagnostic;
            return this;
        };
        Object.defineProperty(RecorderConfigDefault.prototype, "maxJsonStringifyForDiagnostic", {
            get: function () {
                return this._maxJsonStringifyForDiagnostic;
            },
            set: function (value) {
                this._maxJsonStringifyForDiagnostic = value;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(RecorderConfigDefault.prototype, "cacheHandler", {
            get: function () {
                return this._cacheHandler;
            },
            enumerable: true,
            configurable: true
        });
        RecorderConfigDefault.prototype.configCacheHandler = function (value) {
            this._cacheHandler = value;
            return this;
        };
        RecorderConfigDefault.prototype.configCreationIdName = function (creationIdName) { this.creationIdName = creationIdName; return this; };
        RecorderConfigDefault.prototype.configMetadatasName = function (playerMetadatasName) { this.playerMetadatasName = playerMetadatasName; return this; };
        RecorderConfigDefault.prototype.configMaxLazyRefNotificationPerSecond = function (maxLazyRefNotificationPerSecond) { this.maxLazyRefNotificationPerSecond = maxLazyRefNotificationPerSecond; return this; };
        RecorderConfigDefault.prototype.configLazyRefNotificationTimeMeasurement = function (lazyRefNotificationTimeMeasurement) { this.lazyRefNotificationTimeMeasurement = lazyRefNotificationTimeMeasurement; return this; };
        RecorderConfigDefault.prototype.configLazyRefNotificationCountMeasurement = function (value) { this._lazyRefNotificationCountMeasurement = value; return this; };
        Object.defineProperty(RecorderConfigDefault.prototype, "creationIdName", {
            get: function () {
                return this._creationIdName;
            },
            set: function (value) {
                this._creationIdName = value;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(RecorderConfigDefault.prototype, "maxLazyRefNotificationPerSecond", {
            get: function () {
                return this._maxLazyRefNotificationPerSecond;
            },
            set: function (value) {
                this._maxLazyRefNotificationPerSecond = value;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(RecorderConfigDefault.prototype, "lazyRefNotificationTimeMeasurement", {
            get: function () {
                return this._lazyRefNotificationTimeMeasurement;
            },
            set: function (value) {
                this._lazyRefNotificationTimeMeasurement = value;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(RecorderConfigDefault.prototype, "playerMetadatasName", {
            get: function () {
                return this._playerMetadatasName;
            },
            set: function (value) {
                this._playerMetadatasName = value;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(RecorderConfigDefault.prototype, "lazyRefNotificationCountMeasurement", {
            get: function () {
                return this._lazyRefNotificationCountMeasurement;
            },
            set: function (value) {
                this._lazyRefNotificationCountMeasurement = value;
            },
            enumerable: true,
            configurable: true
        });
        RecorderConfigDefault.prototype.configAddFieldProcessors = function (entries) {
            for (var _i = 0, entries_1 = entries; _i < entries_1.length; _i++) {
                var entry = entries_1[_i];
                this._fieldProcessorEntryMap.set(entry.type, entry.processor);
            }
            this._fieldProcessorEntryMap;
            return this;
        };
        RecorderConfigDefault.prototype.getTypeProcessor = function (type) {
            if (this._fieldProcessorEntryMap.get(type)) {
                return this._fieldProcessorEntryMap.get(type);
            }
            else {
                return null;
            }
        };
        Object.defineProperty(RecorderConfigDefault.prototype, "tryReduceLazyRefSubescribersRerun", {
            /**
             * Getter tryReduceLazyRefSubescribersRerun
             * @return {boolean}
             */
            get: function () {
                return this._tryReduceLazyRefSubescribersRerun;
            },
            enumerable: true,
            configurable: true
        });
        RecorderConfigDefault.prototype.configTryReduceLazyRefSubescribersRerun = function (tryReduce) {
            this._tryReduceLazyRefSubescribersRerun = tryReduce;
            return this;
        };
        return RecorderConfigDefault;
    }());
    exports.RecorderConfigDefault = RecorderConfigDefault;
});
//# sourceMappingURL=recorder-config-default.js.map