(function (factory) {
    if (typeof module === "object" && typeof module.exports === "object") {
        var v = factory(require, exports);
        if (v !== undefined) module.exports = v;
    }
    else if (typeof define === "function" && define.amd) {
        define(["require", "exports", "../api/tape", "./recorder-constants", "../api/recorder-config", "./tape-default", "./lodash-like"], factory);
    }
})(function (require, exports) {
    Object.defineProperty(exports, "__esModule", { value: true });
    var tape_1 = require("../api/tape");
    var recorder_constants_1 = require("./recorder-constants");
    var recorder_config_1 = require("../api/recorder-config");
    var tape_default_1 = require("./tape-default");
    var lodash_like_1 = require("./lodash-like");
    var SetCreator = /** @class */ (function () {
        function SetCreator(_session, _refererObj, _refererKey) {
            this._session = _session;
            this._refererObj = _refererObj;
            this._refererKey = _refererKey;
            var thisLocal = this;
            if (!_session) {
                throw new Error('_session can not be null');
            }
            if (!_refererObj) {
                throw new Error('_refererObj can not be null');
            }
            if (!_session) {
                throw new Error('_session can not be null');
            }
            thisLocal.consoleLike = _session.manager.config.getConsole(recorder_config_1.RecorderLogger.SetCreator);
            if (thisLocal.consoleLike.enabledFor(recorder_config_1.RecorderLogLevel.Trace)) {
                thisLocal.consoleLike.group('SetCreator.constructor()');
                thisLocal.consoleLike.debug(_session);
                thisLocal.consoleLike.debug(_refererObj);
                thisLocal.consoleLike.debug(_refererKey);
                thisLocal.consoleLike.groupEnd();
            }
        }
        SetCreator.prototype.createByProxy = function () {
            var _this = this;
            var thisLocal = this;
            var getFunction = function (target, p) {
                if (p) {
                    if (thisLocal.consoleLike.enabledFor(recorder_config_1.RecorderLogLevel.Trace)) {
                        thisLocal.consoleLike.group('SetCreator => getFunction. Intercepting:');
                        thisLocal.consoleLike.debug(target);
                        thisLocal.consoleLike.debug(p);
                        //thisLocal.consoleLike.debug(receiver);
                        thisLocal.consoleLike.groupEnd();
                    }
                    if (p === 'add') {
                        var proxyGet = function (value) {
                            _this.add(target, value);
                            return target.add(value);
                        };
                        return proxyGet;
                    }
                    else if (p === 'delete') {
                        var proxyDelete = function (value) {
                            _this.add(target, value);
                            return target.delete(value);
                        };
                        return proxyDelete;
                    }
                    else if (p === 'clear') {
                        var proxyClear = function () {
                            target.forEach(function (value) {
                                _this.delete(target, value);
                            });
                            target.clear();
                        };
                        return proxyClear;
                    }
                    else if (target[p] && target[p] instanceof Function) {
                        return function (argArray) {
                            return target[p](argArray);
                        };
                    }
                    else {
                        return target[p];
                    }
                }
                else {
                    return undefined;
                }
            };
            return new Proxy(new Set(), {
                get: getFunction
            });
        };
        SetCreator.prototype.add = function (targetSet, value) {
            var propertyOptions = Reflect.getMetadata(recorder_constants_1.RecorderConstants.REFLECT_METADATA_PLAYER_OBJECT_PROPERTY_OPTIONS, this.refererObj, this.refererKey);
            if (!propertyOptions) {
                throw new Error('@RecorderDecorators.property() not defined for ' + this.refererObj.constructor.name + '.' + this.refererKey);
            }
            if (propertyOptions.persistent) {
                var isOnlazyLoad = lodash_like_1.LodashLike.get(targetSet, recorder_constants_1.RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME);
                if (!this.session.isOnRestoreEntireStateFromLiteral() && !isOnlazyLoad) {
                    if (!this.session.isRecording()) {
                        throw new Error('Invalid operation. It is not recording. Is this Error correct?!');
                    }
                    var allMD = this.session.resolveMetadatas({ object: value, refererObject: this.refererObj, key: this.refererKey });
                    var mdRefererObj = allMD.refererObjMd;
                    var mdValue = allMD.objectMd;
                    //recording tape
                    var action = new tape_default_1.TapeActionDefault();
                    action.fieldName = this.refererKey;
                    action.actionType = tape_1.TapeActionType.CollectionAdd;
                    if (mdRefererObj.$signature$) {
                        action.ownerSignatureStr = mdRefererObj.$signature$;
                    }
                    else if (lodash_like_1.LodashLike.has(this.refererObj, this.session.manager.config.creationIdName)) {
                        action.ownerCreationRefId = lodash_like_1.LodashLike.get(this.refererObj, this.session.manager.config.creationIdName);
                    }
                    else if (!mdRefererObj.$isComponentPlayerObjectId$) {
                        throw new Error('The proprerty \'' + this.refererKey + ' from \'' + this.refererObj.constructor.name + '\' has a not managed owner');
                    }
                    if (value != null) {
                        if (mdValue.$signature$) {
                            action.settedSignatureStr = mdValue.$signature$;
                        }
                        else if (lodash_like_1.LodashLike.has(value, this.session.manager.config.creationIdName)) {
                            action.settedCreationRefId = lodash_like_1.LodashLike.get(value, this.session.manager.config.creationIdName);
                        }
                        else {
                            throw new Error('The proprerty \'' + this.refererKey + ' of \'' + this.refererObj.constructor.name + '\'.  value not managed owner: \'' + value.constructor.name + '\'');
                        }
                    }
                    this.session.addTapeAction(action);
                }
            }
        };
        SetCreator.prototype.delete = function (targetSet, value) {
            var propertyOptions = Reflect.getMetadata(recorder_constants_1.RecorderConstants.REFLECT_METADATA_PLAYER_OBJECT_PROPERTY_OPTIONS, this.refererObj, this.refererKey);
            if (!propertyOptions) {
                throw new Error('@RecorderDecorators.property() not defined for ' + this.refererObj.constructor.name + '.' + this.refererKey);
            }
            if (propertyOptions.persistent) {
                var isOnlazyLoad = lodash_like_1.LodashLike.get(targetSet, recorder_constants_1.RecorderConstants.ENTITY_IS_ON_LAZY_LOAD_NAME);
                if (!this.session.isOnRestoreEntireStateFromLiteral() && !isOnlazyLoad) {
                    if (!this.session.isRecording()) {
                        throw new Error('Invalid operation. It is not recording. Is this Error correct?!');
                    }
                    var allMD = this.session.resolveMetadatas({ object: value, refererObject: this.refererObj, key: this.refererKey });
                    var mdRefererObj = allMD.refererObjMd;
                    var mdValue = allMD.objectMd;
                    //recording tape
                    var action = new tape_default_1.TapeActionDefault();
                    action.fieldName = this.refererKey;
                    action.actionType = tape_1.TapeActionType.CollectionRemove;
                    if (mdRefererObj.$signature$) {
                        action.ownerSignatureStr = mdRefererObj.$signature$;
                    }
                    else if (lodash_like_1.LodashLike.has(this.refererObj, this.session.manager.config.creationIdName)) {
                        action.ownerCreationRefId = lodash_like_1.LodashLike.get(this.refererObj, this.session.manager.config.creationIdName);
                    }
                    else {
                        throw new Error('The proprerty \'' + this.refererKey + ' of \'' + this.refererObj.constructor + '\' has a not managed owner');
                    }
                    if (value != null) {
                        if (mdValue.$signature$) {
                            action.settedSignatureStr = mdValue.$signature$;
                        }
                        else if (lodash_like_1.LodashLike.has(value, this.session.manager.config.creationIdName)) {
                            action.settedCreationRefId = lodash_like_1.LodashLike.get(value, this.session.manager.config.creationIdName);
                        }
                        else {
                            throw new Error('The proprerty \'' + this.refererKey + ' of \'' + this.refererObj.constructor + '\'. not managed value: \'' + value.constructor.name + '\'');
                        }
                    }
                    this.session.addTapeAction(action);
                }
            }
        };
        Object.defineProperty(SetCreator.prototype, "session", {
            /**
             * Getter session
             * @return {RecorderSession}
             */
            get: function () {
                return this._session;
            },
            /**
             * Setter session
             * @param {RecorderSession} value
             */
            set: function (value) {
                this._session = value;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(SetCreator.prototype, "refererObj", {
            /**
             * Getter refererObj
             * @return {any}
             */
            get: function () {
                return this._refererObj;
            },
            /**
             * Setter refererObj
             * @param {any} value
             */
            set: function (value) {
                this._refererObj = value;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(SetCreator.prototype, "refererKey", {
            /**
             * Getter refererKey
             * @return {string}
             */
            get: function () {
                return this._refererKey;
            },
            /**
             * Setter refererKey
             * @param {string} value
             */
            set: function (value) {
                this._refererKey = value;
            },
            enumerable: true,
            configurable: true
        });
        return SetCreator;
    }());
    exports.SetCreator = SetCreator;
});
//# sourceMappingURL=set-creator.js.map